from .base import EvalContext, EvalResult
from .functions import Function, function


@function("AND")
class AndFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        for arg in self.args:
            if not await arg.eval(context):
                return 0

        return 1


@function("OR")
class OrFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        for arg in self.args:
            if await arg.eval(context):
                return 1

        return 0


@function("NOT")
class NotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def _eval(self, context: EvalContext) -> EvalResult:
        return int(not bool((await self.eval_args(context))[0]))


@function("XOR")
class XOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        e1 = bool(eval_args[0])
        e2 = bool(eval_args[1])

        return int(e1 and not e2 or e2 and not e1)



from typing import Any, TypeAlias


PortValue: TypeAlias = int | float | bool
NullablePortValue: TypeAlias = int | float | bool | None
PortValueChoices: TypeAlias = list[dict[str, str | int | float]]

Attribute: TypeAlias = int | float | bool | str | list[dict] | None
Attributes: TypeAlias = dict[str, Attribute]

AttributeDefinition: TypeAlias = dict[str, Any]
AttributeDefinitions: TypeAlias = dict[str, AttributeDefinition]

GenericJSONDict: TypeAlias = dict[str, Any]
GenericJSONList: TypeAlias = list[dict[str, Any]]



import logging
import re

from collections.abc import Iterable
from typing import Any

import bson
import pymongo.database
import pymongo.errors

from qtoggleserver.persist import BaseDriver
from qtoggleserver.persist.typing import Id, Record


logger = logging.getLogger(__name__)

_OBJECT_ID_RE = re.compile("^[0-9a-f]{24}$")
DEFAULT_DB = "qtoggleserver"

FILTER_OP_MAPPING = {"gt": "$gt", "ge": "$gte", "lt": "$lt", "le": "$lte", "in": "$in"}


class MongoDriver(BaseDriver):
    def __init__(self, host: str = "127.0.0.1", port: int = 27017, db: str = DEFAULT_DB, **kwargs) -> None:
        self._host: str = host
        self._port: int = port
        self._db_name: str = db

        self._client: pymongo.MongoClient | None = None
        self._db: pymongo.database.Database | None = None

    async def init(self) -> None:
        logger.debug("connecting to %s:%s/%s", self._host, self._port, self._db_name)
        self._client = pymongo.MongoClient(
            self._host, self._port, serverSelectionTimeoutMS=200, connectTimeoutMS=200
        )
        self._db = self._client[self._db_name]

    async def cleanup(self) -> None:
        logger.debug("disconnecting mongo client")

        self._client.close()

    async def query(
        self,
        collection: str,
        fields: list[str] | None,
        filt: dict[str, Any],
        sort: list[tuple[str, bool]],
        limit: int | None,
    ) -> Iterable[Record]:
        if fields:
            fields = {f: 1 for f in fields}
            if "id" in fields:
                fields["_id"] = 1
            else:
                fields["_id"] = 0

        if "id" in filt:
            filt = dict(filt)
            filt["_id"] = self._id_to_db_rec(filt.pop("id"))

        db_filt = self._filt_to_db(filt)

        q = self._db[collection].find(db_filt, fields)

        if len(sort) > 0:
            sort = [(f, [pymongo.ASCENDING, pymongo.DESCENDING][r]) for f, r in sort]
            q = q.sort(sort)

        if limit is not None:
            q = q.limit(limit)

        return self._query_gen_wrapper(q)

    async def insert(self, collection: str, record: Record) -> Id:
        record = dict(record)
        if "id" in record:
            record["_id"] = self._id_to_db(record.pop("id"))

        return self._id_from_db(self._db[collection].insert_one(record).inserted_id)

    async def update(self, collection: str, record_part: Record, filt: dict[str, Any]) -> int:
        if "id" in record_part:
            record_part = dict(record_part)
            record_part["_id"] = self._id_to_db(record_part.pop("id"))

        if "id" in filt:
            filt = dict(filt)
            filt["_id"] = self._id_to_db_rec(filt.pop("id"))

        db_filt = self._filt_to_db(filt)

        return self._db[collection].update_many(db_filt, {"$set": record_part}, upsert=False).modified_count

    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        record = dict(record)
        id_ = self._id_to_db(id_)
        record["_id"] = id_

        matched = self._db[collection].replace_one({"_id": id_}, record, upsert=False).matched_count

        return matched > 0

    async def remove(self, collection: str, filt: dict[str, Any]) -> int:
        if "id" in filt:
            filt = dict(filt)
            filt["_id"] = self._id_to_db_rec(filt.pop("id"))

        db_filt = self._filt_to_db(filt)

        return self._db[collection].delete_many(db_filt).deleted_count

    def is_samples_supported(self) -> bool:
        return True

    async def ensure_index(self, collection: str, index: list[tuple[str, bool]] | None) -> None:
        if index:
            index = [(f, [pymongo.ASCENDING, pymongo.DESCENDING][r]) for f, r in index]
        else:  # assuming samples collection
            index = [("ts", pymongo.ASCENDING)]

        try:
            self._db[collection].create_index(index)
        except pymongo.errors.DuplicateKeyError:
            pass

    @classmethod
    def _query_gen_wrapper(cls, q: Iterable[Record]) -> Iterable[Record]:
        for r in q:
            if "_id" in r:
                r["id"] = cls._id_from_db(r.pop("_id"))

            yield r

    @staticmethod
    def _id_to_db(id_: str) -> str | bson.ObjectId:
        if isinstance(id_, str) and _OBJECT_ID_RE.match(id_):
            return bson.ObjectId(id_)

        else:
            return id_

    @staticmethod
    def _id_from_db(id_: str | bson.ObjectId) -> str:
        if isinstance(id_, bson.ObjectId):
            return str(id_)

        else:
            return id_

    def _id_to_db_rec(self, obj: dict | list | str) -> dict | list | bson.ObjectId:
        if isinstance(obj, dict):
            return {key: self._id_to_db_rec(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [self._id_to_db(value) for value in obj]

        else:  # assuming str
            return self._id_to_db(obj)

    @staticmethod
    def _filt_to_db(filt: dict[str, Any]) -> dict[str, Any]:
        db_filt = {}
        for key, value in filt.items():
            if isinstance(value, dict):  # filter with operators
                value = {FILTER_OP_MAPPING[k]: v for k, v in value.items()}

            db_filt[key] = value

        return db_filt



from .base import EvalContext, EvalResult
from .functions import Function, function


@function("TIME")
class TimeFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        return context.timestamp


@function("TIMEMS")
class TimeMSFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = {"asap"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        return context.now_ms



import logging
import threading

from collections.abc import Iterable
from typing import Any

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils
from qtoggleserver.utils import dynload as dynload_utils
from qtoggleserver.utils import json as json_utils

from .base import BaseDriver
from .typing import Id, Record, Sample, SampleValue


logger = logging.getLogger(__name__)

_thread_local: threading.local = threading.local()


async def _get_driver() -> BaseDriver:
    if not hasattr(_thread_local, "driver"):
        driver_args = conf_utils.obj_to_dict(settings.persist)
        driver_class_path = driver_args.pop("driver")

        try:
            logger.debug("loading persistence driver %s", driver_class_path)
            driver_class = dynload_utils.load_attr(driver_class_path)
            driver = driver_class(**driver_args)
            await driver.init()
            _thread_local.driver = driver
        except Exception as e:
            logger.error("failed to load persistence driver %s: %s", driver_class_path, e, exc_info=True)

            raise

    return _thread_local.driver


async def query(
    collection: str,
    fields: list[str] | None = None,
    filt: dict[str, Any] | None = None,
    sort: str | list[str] | None = None,
    limit: int | None = None,
) -> Iterable[Record]:
    """Return records from `collection`.

    Optionally project the returned records by only returning given `fields`.

    Filter records according to `filt`. Besides filtering by exact field value, the following filter operators are
    supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

    Sort results according to `sort` argument which is a list of fields optionally preceded by `-` for descending sort.

    Optionally limit the number of returned records to `limit`."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "querying %s (%s) where %s (sort=%s, limit=%s)",
            collection,
            json_utils.dumps(fields) if fields else "all fields",
            json_utils.dumps(filt, extra_types=json_utils.EXTRA_TYPES_EXTENDED),
            json_utils.dumps(sort),
            json_utils.dumps(limit),
        )

    filt = filt or {}
    sort = sort or []
    if isinstance(sort, str):
        sort = [sort]

    # Transform '-field' into (field, descending)
    sort_tuples = [(s[1:], True) if s.startswith("-") else (s, False) for s in sort]

    driver = await _get_driver()
    return await driver.query(collection, fields, filt, sort_tuples, limit)


async def get(collection: str, id_: Id) -> Record | None:
    """Return the record with a given `id` from `collection`.

    If no such record is found, `None` is returned."""

    logger.debug("getting record with id %s from %s", id_, collection)

    driver = await _get_driver()
    records = list(await driver.query(collection, fields=None, filt={"id": id_}, sort=[], limit=1))
    if len(records) > 1:
        logger.warning("more than one record with same id %s found in collection %s", id_, collection)

    if records:
        return records[0]

    return None


async def get_value(name: str, default: Any | None = None) -> Any:
    """Return the value of object with name `name` by querying the collection with the same name and considering the
    first (and only) record.

    If no such record is found, `default` is returned."""

    logger.debug("getting value of %s", name)

    driver = await _get_driver()
    records = list(await driver.query(name, fields=None, filt={}, sort=[], limit=2))
    if len(records) > 1:
        logger.warning("more than one record found in single-value collection %s", name)
        record = records[0]
    elif len(records) > 0:
        record = records[0]
    else:
        return default

    return record["value"]


async def set_value(name: str, value: Any) -> None:
    """Set the value of object with name `name` by inserting into or updating the collection with the same name and
    considering the fist (and only) record."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug("setting %s to %s", name, json_utils.dumps(value, extra_types=json_utils.EXTRA_TYPES_EXTENDED))

    driver = await _get_driver()
    record = {"value": value}

    records = list(await driver.query(name, fields=["id"], filt={}, sort=[], limit=2))
    if len(records) > 1:
        logger.warning("more than one record found in single-value collection %s", name)

        id_ = records[0]["id"]
    elif len(records) > 0:
        id_ = records[0]["id"]

    else:
        id_ = None

    if id_ is None:
        await driver.insert(name, record)

    else:
        await driver.replace(name, id_, record)


async def insert(collection: str, record: Record) -> Id:
    """Insert `record` into `collection`.

    Return the associated record ID."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "inserting %s into %s", json_utils.dumps(record, extra_types=json_utils.EXTRA_TYPES_EXTENDED), collection
        )

    driver = await _get_driver()
    return await driver.insert(collection, record)


async def update(collection: str, record_part: Record, filt: dict[str, Any] | None = None) -> int:
    """Update records from `collection` with fields given in `record_part`.

    Only records that match the `filt` filter will be updated. Besides filtering by exact field value, the following
    filter operators are supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

    Return the total number of records that were updated."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "updating %s where %s with %s",
            collection,
            json_utils.dumps(filt or {}, extra_types=json_utils.EXTRA_TYPES_EXTENDED),
            json_utils.dumps(record_part, extra_types=json_utils.EXTRA_TYPES_EXTENDED),
        )

    driver = await _get_driver()
    count = await driver.update(collection, record_part, filt or {})

    logger.debug("modified %s records in %s", count, collection)

    return count


async def replace(collection: str, id_: Id, record: Record) -> bool:
    """Replace record with `id` in `collection`.

    Return `True` if matched and replaced."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "replacing record with id %s with %s in %s",
            id_,
            json_utils.dumps(record, extra_types=json_utils.EXTRA_TYPES_EXTENDED),
            collection,
        )

    record = dict(record, id=id_)  # make sure the new record contains the id field
    driver = await _get_driver()
    replaced = await driver.replace(collection, id_, record)
    if replaced:
        logger.debug("replaced record with id %s in %s", id_, collection)

        return False

    else:
        await driver.insert(collection, dict(record, id=id_))
        logger.debug("inserted record with id %s in %s", id_, collection)

        return True


async def remove(collection: str, filt: dict[str, Any] | None = None) -> int:
    """Remove records from `collection`.

    Only records that match the `filt` filter will be removed. Besides filtering by exact field value, the following
    filter operators are supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

    Return the total number of records that were removed."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "removing from %s where %s",
            collection,
            json_utils.dumps(filt or {}, extra_types=json_utils.EXTRA_TYPES_EXTENDED),
        )

    driver = await _get_driver()
    count = await driver.remove(collection, filt or {})

    logger.debug("removed %s records from %s", count, collection)

    return count


async def get_samples_slice(
    collection: str,
    obj_id: Id,
    from_timestamp: int | None = None,
    to_timestamp: int | None = None,
    limit: int | None = None,
    sort_desc: bool = False,
) -> Iterable[Sample]:
    """Return the samples of `obj_id` from `collection`.

    Filter results by an interval of time, if `from_timestamp` and/or `to_timestamp` are not `None`.
    `from_timestamp` is inclusive, while `to_timestamp` is exclusive.

    Optionally limit results to `limit` number of records, if not `None`.

    Sort the results by timestamp according to the value of `sort_desc`."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "getting samples of object %s from %s between %s and %s (sort=%s, limit=%s)",
            obj_id,
            collection,
            json_utils.dumps(from_timestamp),
            json_utils.dumps(to_timestamp),
            ["asc", "desc"][sort_desc],
            json_utils.dumps(limit),
        )

    driver = await _get_driver()
    return await driver.get_samples_slice(collection, obj_id, from_timestamp, to_timestamp, limit, sort_desc)


async def get_samples_by_timestamp(collection: str, obj_id: Id, timestamps: list[int]) -> Iterable[SampleValue]:
    """For each timestamp in `timestamps`, return the sample of `obj_id` from `collection` that was saved right
    before the (or at the exact) timestamp.
    """

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "getting samples of object %s from %s for %d timestamps",
            obj_id,
            collection,
            len(timestamps),
        )

    driver = await _get_driver()
    return await driver.get_samples_by_timestamp(collection, obj_id, timestamps)


async def save_sample(collection: str, obj_id: Id, timestamp: int, value: SampleValue) -> None:
    """Save a sample of an object with `obj_id` at a given `timestamp` to a specified `collection`."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "saving sample value %s of object %s at %s into %s",
            json_utils.dumps(value),
            obj_id,
            timestamp,
            collection,
        )

    driver = await _get_driver()
    return await driver.save_sample(collection, obj_id, timestamp, value)


async def remove_samples(
    collection: str,
    obj_ids: list[Id] | None = None,
    from_timestamp: int | None = None,
    to_timestamp: int | None = None,
) -> int:
    """Remove samples from `collection`.

    If `obj_ids` is not `None`, only remove samples of given object ids.

    If `from_timestamp` and/or `to_timestamp` are not `None`, only remove samples within specified interval of time.
    `from_timestamp` is inclusive, while `to_timestamp` is exclusive.

    Return the number of removed samples."""

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            "removing samples of objects (%s) between %s and %s from %s",
            ", ".join(obj_ids) if obj_ids else "all objects",
            json_utils.dumps(from_timestamp),
            json_utils.dumps(to_timestamp),
            collection,
        )

    driver = await _get_driver()
    count = await driver.remove_samples(collection, obj_ids, from_timestamp, to_timestamp)
    logger.debug("removed %s samples", count, collection)

    return count


def is_samples_supported() -> bool:
    """Tell whether samples are supported by the current persistence driver or not."""

    # We need this function to *not* be async, therefore we try to obtain a reference to the existing driver rather than
    # calling the async function `_get_driver()`. We rely on the fact that it will always be called after driver
    # initialization and thus the `_thread_local` variable will have the `driver` attribute set.
    driver = getattr(_thread_local, "driver", None)
    if not driver:
        return False

    return driver.is_samples_supported()


async def ensure_index(collection: str, index: str | list[str] | None = None) -> None:
    """Create an index on `collection` if not already present.

    `index` is a list of field names optionally preceded by a `-` sign for a descending index.

    If `index` is `None`, the collection is assumed to be a collection of samples where the index is considered to
    be ascending on the timestamp field."""

    if isinstance(index, str):
        index = [index]

    # Transform '-field' into (field, descending)
    index_tuples = []
    if index:
        index_tuples = [(i[1:], True) if i.startswith("-") else (i, False) for i in index]

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug("ensuring index %s in %s", json_utils.dumps(index_tuples), collection)

    driver = await _get_driver()
    await driver.ensure_index(collection, index_tuples)


async def init() -> None:
    """Initialize the persistence subsystem."""

    driver = await _get_driver()

    # Do a dummy query so that if there's any problem in querying the collection, an exception is raised now.
    await driver.query("device", fields=None, filt={}, sort=[], limit=1)


async def cleanup() -> None:
    """Clean up the persistence subsystem."""

    driver = await _get_driver()
    await driver.cleanup()
    _thread_local.driver = None



class PeripheralException(Exception):
    pass


class NoSuchDriver(PeripheralException):
    pass


class DuplicatePeripheral(PeripheralException):
    pass


class NotOurPort(PeripheralException):
    pass



import asyncio
import copy
import hashlib
import inspect
import logging
import re
import socket
import sys
import time

from collections.abc import Callable
from datetime import datetime, timezone
from typing import Any

from qtoggleserver import system, version
from qtoggleserver.conf import settings
from qtoggleserver.core.typing import AttributeDefinitions, Attributes, GenericJSONDict
from qtoggleserver.system import fwupdate
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd

from . import events as device_events


logger = logging.getLogger(__name__)


EMPTY_PASSWORD_HASH = hashlib.sha256(b"").hexdigest()
NETWORK_ATTRS_WATCH_INTERVAL = 5
ATTRDEFS_CACHE_TIMEOUT = 10
ATTRDEF_CALLABLE_FIELDS = {"modifiable", "min", "max", "enabled"}


name: str = re.sub(r"[^a-zA-Z0-9_-]", "", socket.gethostname())
if not re.match("^[a-zA-Z_]", name):  # make sure name starts with a letter or underscore
    name = f"q{name}"
name = name[:32]

display_name: str = ""
admin_password_hash: str | None = None
normal_password_hash: str | None = None
viewonly_password_hash: str | None = None

_schema: GenericJSONDict | None = None
_attrdefs: AttributeDefinitions | None = None
_attrs_watch_task: asyncio.Task | None = None


def attr_get_name() -> str:
    global name
    if settings.core.device_name.get_cmd:
        result = run_get_cmd(settings.core.device_name.get_cmd, cmd_name="device name", required_fields=["name"])
        name = result["name"]

    return name


def attr_set_name(value: str) -> None:
    global name
    name = value

    if settings.core.device_name.set_cmd:
        run_set_cmd(settings.core.device_name.set_cmd, cmd_name="device name", name=value)


def attr_get_display_name() -> str:
    return display_name


def attr_set_display_name(value: str) -> None:
    global display_name
    display_name = value


def attr_get_api_version() -> str:
    from qtoggleserver.core import api as core_api

    return core_api.API_VERSION


async def attr_get_version() -> str:
    if settings.system.fwupdate.driver:
        return await fwupdate.get_current_version()
    else:
        return version.VERSION


def attr_get_flags() -> list[str]:
    from qtoggleserver.core import history as core_history

    flags = ["expressions"]
    if settings.system.fwupdate.driver:
        flags.append("firmware")

    if settings.core.backup_support:
        flags.append("backup")

    if core_history.is_enabled():
        flags.append("history")

    if settings.core.listen_support:
        flags.append("listen")

    if settings.slaves.enabled:
        flags.append("master")

    if settings.reverse.enabled:
        flags.append("reverse")

    if settings.core.sequences_support:
        flags.append("sequences")

    if settings.core.tls_support:
        flags.append("tls")

    if settings.webhooks.enabled:
        flags.append("webhooks")

    return flags


def attr_get_password(which: str) -> str:
    core_device_attrs = sys.modules[__name__]
    password_hash = getattr(core_device_attrs, f"{which}_password_hash")
    return ["set", ""][password_hash == EMPTY_PASSWORD_HASH]


def attr_set_password(which: str, value: str) -> None:
    # Call password set command, if available
    if settings.core.passwords.set_cmd:
        run_set_cmd(
            settings.core.passwords.set_cmd, cmd_name="password", log_values=False, username=which, password=value
        )

    core_device_attrs = sys.modules[__name__]
    password_hash = hashlib.sha256(value.encode()).hexdigest()
    setattr(core_device_attrs, f"{which}_password_hash", password_hash)


ATTRDEFS = {
    "name": {
        "type": "string",
        "modifiable": True,
        "min": 1,
        "max": 32,
        "pattern": r"^[_a-zA-Z][_a-zA-Z0-9-]{0,31}$",
        "standard": True,
        "persisted": True,
        "getter": attr_get_name,
        "setter": attr_set_name,
    },
    "display_name": {
        "type": "string",
        "modifiable": True,
        "max": 64,
        "standard": True,
        "persisted": True,
        "getter": attr_get_display_name,
        "setter": attr_set_display_name,
    },
    "version": {
        "type": "string",
        "standard": True,
        "getter": attr_get_version,
    },
    "firmware_auto_update": {
        "type": "boolean",
        "modifiable": True,
        "standard": True,
        "enabled": lambda: bool(settings.system.fwupdate.driver),
        "getter": fwupdate.is_auto_update_enabled,
        "setter": fwupdate.set_auto_update_enabled,
    },
    "api_version": {
        "type": "string",
        "standard": True,
        "getter": attr_get_api_version,
    },
    "vendor": {
        "type": "string",
        "standard": True,
        "getter": lambda: version.VENDOR,
    },
    "admin_password": {
        "type": "string",
        "modifiable": True,
        "max": 32,
        "standard": True,
        "getter": lambda: attr_get_password("admin"),
        "setter": lambda v: attr_set_password("admin", v),
    },
    "normal_password": {
        "type": "string",
        "modifiable": True,
        "max": 32,
        "standard": True,
        "getter": lambda: attr_get_password("normal"),
        "setter": lambda v: attr_set_password("normal", v),
    },
    "viewonly_password": {
        "type": "string",
        "modifiable": True,
        "max": 32,
        "standard": True,
        "getter": lambda: attr_get_password("viewonly"),
        "setter": lambda v: attr_set_password("viewonly", v),
    },
    "flags": {
        "type": ["string"],
        "standard": True,
        "getter": attr_get_flags,
    },
    "virtual_ports": {
        "type": "number",
        "standard": True,
        "enabled": lambda: bool(settings.core.virtual_ports),
        "getter": lambda: settings.core.virtual_ports,
    },
    "uptime": {
        "type": "number",
        "standard": True,
        "getter": system.uptime,
    },
    "date": {
        "type": "number",
        "modifiable": system.date.has_set_date_support,
        # mark as non-standard if no set support, to make attribute non-modifiable
        "standard": system.date.has_set_date_support,
        "enabled": system.date.has_real_date_time,
        "getter": lambda: int(time.time()),
        "setter": lambda v: system.date.set_date(datetime.fromtimestamp(v, tz=timezone.utc)),
    },
    "timezone": {
        "type": "string",
        "modifiable": True,
        "choices": [{"value": zone} for zone in system.date.get_timezones()],
        "standard": True,
        "enabled": system.date.has_timezone_support,
        "getter": system.date.get_timezone,
        "setter": system.date.set_timezone,
    },
    "wifi_ssid": {
        "type": "string",
        "max": 32,
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_wifi_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_wifi_config,
            "key": "ssid",
        },
        "setter": {
            "call": system.net.set_wifi_config,
            "key": "ssid",
        },
    },
    "wifi_key": {
        "type": "string",
        "max": 64,
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_wifi_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_wifi_config,
            "key": "psk",
        },
        "setter": {
            "call": system.net.set_wifi_config,
            "key": "psk",
        },
    },
    "wifi_bssid": {
        "type": "string",
        "pattern": r"^([a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2})?$",
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_wifi_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_wifi_config,
            "key": "bssid",
        },
        "setter": {
            "call": system.net.set_wifi_config,
            "key": "bssid",
        },
    },
    "wifi_bssid_current": {
        "type": "string",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_wifi_support,
        "getter": {
            "call": system.net.get_wifi_config,
            "key": "bssid_current",
        },
    },
    "wifi_signal_strength": {
        "type": "number",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_wifi_support,
        "getter": {
            "call": system.net.get_wifi_config,
            "key": "signal_strength_current",
            "transform": lambda v: int(v),
        },
    },
    "ip_address": {
        "type": "string",
        "pattern": r"^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$",
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "address",
        },
        "setter": {
            "call": system.net.set_ip_config,
            "key": "address",
        },
    },
    "ip_netmask": {
        "type": "number",
        "min": 0,
        "max": 31,
        "integer": True,
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "reboot": True,
        "getter": {"call": system.net.get_ip_config, "key": "netmask", "transform": lambda v: v or 0},
        "setter": {"call": system.net.set_ip_config, "key": "netmask", "transform": lambda v: str(v)},
    },
    "ip_gateway": {
        "type": "string",
        "pattern": r"^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$",
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "gateway",
        },
        "setter": {
            "call": system.net.set_ip_config,
            "key": "gateway",
        },
    },
    "ip_dns": {
        "type": "string",
        "pattern": r"^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$",
        "modifiable": True,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "reboot": True,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "dns",
        },
        "setter": {
            "call": system.net.set_ip_config,
            "key": "dns",
        },
    },
    "ip_address_current": {
        "type": "string",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "address_current",
        },
    },
    "ip_netmask_current": {
        "type": "number",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "getter": {"call": system.net.get_ip_config, "key": "netmask_current", "transform": lambda v: v or 0},
    },
    "ip_gateway_current": {
        "type": "string",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "gateway_current",
        },
    },
    "ip_dns_current": {
        "type": "string",
        "modifiable": False,
        "standard": True,
        "enabled": system.net.has_ip_support,
        "getter": {
            "call": system.net.get_ip_config,
            "key": "dns_current",
        },
    },
    "cpu_usage": {
        "type": "number",
        "min": 0,
        "max": 100,
        "modifiable": False,
        "standard": True,
        "getter": system.get_cpu_usage,
    },
    "mem_usage": {
        "type": "number",
        "min": 0,
        "max": 100,
        "modifiable": False,
        "standard": True,
        "getter": system.get_mem_usage,
    },
    "storage_usage": {
        "type": "number",
        "min": 0,
        "max": 100,
        "modifiable": False,
        "standard": True,
        "enabled": system.storage.has_storage_support,
        "getter": system.storage.get_storage_usage,
    },
    "temperature": {
        "type": "number",
        "modifiable": False,
        "min": lambda: settings.system.temperature.min,
        "max": lambda: settings.system.temperature.max,
        # mark as non-standard to expose min/max fields
        "standard": False,
        "enabled": system.temperature.has_temperature_support,
        "getter": system.temperature.get_temperature,
    },
    "battery_level": {
        "type": "number",
        "min": 0,
        "max": 100,
        "modifiable": False,
        "standard": True,
        "enabled": system.battery.has_battery_support,
        "getter": system.battery.get_battery_level,
    },
}


class DeviceAttributeError(Exception):
    def __init__(self, error: str, attribute: str) -> None:
        self.error: str = error
        self.attribute: str = attribute


def get_attrdefs() -> AttributeDefinitions:
    global _attrdefs

    if _attrdefs is None:
        logger.debug("initializing attribute definitions")
        _attrdefs = copy.deepcopy(ATTRDEFS)

        # Transform some callable values into corresponding results
        for n, attrdef in list(_attrdefs.items()):
            for k, v in attrdef.items():
                if callable(v) and k in ATTRDEF_CALLABLE_FIELDS:
                    attrdef[k] = v()

            if attrdef.pop("enabled", True) is False:
                _attrdefs.pop(n)

    return _attrdefs


def get_schema(loose: bool = False) -> GenericJSONDict:
    global _schema

    # Use cached value, but only when loose is false, as loose schema is never cached
    if _schema is not None and not loose:
        return _schema

    schema = {"type": "object", "properties": {}, "additionalProperties": loose}

    for n, attrdef in get_attrdefs().items():
        if not attrdef.get("modifiable"):
            continue

        attr_schema = dict(attrdef)
        if attr_schema["type"] == "string":
            if "min" in attr_schema:
                attr_schema["minLength"] = attr_schema.pop("min")

            if "max" in attr_schema:
                attr_schema["maxLength"] = attr_schema.pop("max")
        elif attr_schema["type"] == "number":
            if attr_schema.get("integer"):
                attr_schema["type"] = "integer"

            if "min" in attr_schema:
                attr_schema["minimum"] = attr_schema.pop("min")

            if "max" in attr_schema:
                attr_schema["maximum"] = attr_schema.pop("max")

        if "choices" in attrdef:
            attr_schema["enum"] = [c["value"] for c in attr_schema.pop("choices")]

        attr_schema.pop("modifiable", None)
        attr_schema.pop("standard", None)
        attr_schema.pop("persisted", None)
        attr_schema.pop("getter", None)
        attr_schema.pop("setter", None)
        attr_schema.pop("reboot", None)

        schema["properties"][n] = attr_schema

    if not loose:
        _schema = schema

    return schema


async def get_attrs() -> Attributes:
    attrdefs = get_attrdefs()

    # Do a first round to gather all required calls and ensure we only call each function once, caching its result
    call_results = {}
    for n, attrdef in attrdefs.items():
        getter = attrdef["getter"]
        if isinstance(getter, dict):
            call = getter["call"]
            call_results[call] = None

    for call in call_results.keys():
        result = call()
        if inspect.isawaitable(result):
            result = await result
        call_results[call] = result

    # Do a second round to prepare attribute values
    attrs = {}
    for n, attrdef in attrdefs.items():
        getter = attrdef["getter"]
        if not getter:
            continue

        if isinstance(getter, dict):
            call = getter["call"]
            key = getter.get("key")
            transform = getter.get("transform")
            value = call_results[call]
            if key:
                value = value[key]
            if transform:
                value = transform(value)
        elif callable(getter):
            value = getter()
            if inspect.isawaitable(value):
                value = await value
        else:
            continue

        attrs[n] = value

    return attrs


async def set_attrs(attrs: Attributes, ignore_extra: bool = False) -> bool:
    core_device_attrs = sys.modules[__name__]

    reboot_required = False
    attrdefs = get_attrdefs()

    # Call getters first so that we have all current attribute values, needed for a complete setter argument list
    getter_call_results = {}
    for n, attrdef in attrdefs.items():
        getter = attrdef["getter"]
        if isinstance(getter, dict):
            call = getter["call"]
            getter_call_results[call] = None

    for call in getter_call_results.keys():
        result = call()
        if inspect.isawaitable(result):
            result = await result
        getter_call_results[call] = result

    call_values: dict[Callable, dict] = {}

    for n, value in attrs.items():
        # A few attributes may carry sensitive information, so treat them separately and do not log their values
        if n.count("password") or n == "wifi_key":
            logger.debug("setting device attribute %s", n)
        else:
            logger.debug("setting device attribute %s = %s", n, json_utils.dumps(value))

        try:
            attrdef = attrdefs[n]
        except KeyError:
            if ignore_extra:
                continue
            else:
                raise DeviceAttributeError("no-such-attribute", n) from None

        if not attrdef.get("modifiable"):
            if ignore_extra:
                continue
            else:
                raise DeviceAttributeError("attribute-not-modifiable", n)

        reboot_required = reboot_required or attrdef.get("reboot", False)

        # Used by `PUT /device` API call to restore passwords stored as hashes
        if n.endswith("_password_hash") and hasattr(core_device_attrs, n):
            # FIXME: Password set command cannot be called with hash, and we don't have clear-text password here.
            #        A solution would be to use sha256 crypt algorithm w/o salt for Unix password (watch for the special
            #        alphabet and for number of rounds defaulting to 5000)
            setattr(core_device_attrs, n, value)
            continue

        setter = attrdef["setter"]
        if isinstance(setter, dict):
            call = setter["call"]
            key = setter.get("key")
            transform = setter.get("transform")
            if transform:
                value = transform(value)
            if key:
                call_values.setdefault(call, {})[key] = value
            else:
                result = call(value)
                if inspect.isawaitable(result):
                    await result
        else:
            result = setter(value)
            if inspect.isawaitable(result):
                await result

    # Actually do the calls with corresponding values
    for call, values in call_values.items():
        # Fill in any missing setter parameters using corresponding getter values
        for n, attrdef in attrdefs.items():
            getter = attrdef["getter"]
            setter = attrdef.get("setter")
            if not isinstance(getter, dict) or not isinstance(setter, dict) or setter["call"] is not call:
                continue
            call_result = getter_call_results.get(getter["call"])
            if not call_result:
                continue
            values.setdefault(setter["key"], call_result[setter["key"]])

        result = call(**values)
        if inspect.isawaitable(result):
            await result

    return reboot_required


async def to_json() -> GenericJSONDict:
    attrdefs: AttributeDefinitions = copy.deepcopy(get_attrdefs())
    filtered_attrdefs: AttributeDefinitions = {}
    for attr_name, attrdef in attrdefs.items():
        if attrdef.pop("standard", False):
            continue

        # Remove unwanted fields from attribute definition

        attrdef.pop("reboot", None)
        attrdef.pop("persisted", None)
        attrdef.pop("pattern", None)  # TODO: remove this line once pattern becomes an API-defined attrdef field
        attrdef.pop("setter", None)
        attrdef.pop("getter", None)

        for key in list(attrdef):
            if key.startswith("_"):
                attrdef.pop(key)

        filtered_attrdefs[attr_name] = attrdef

    result: dict[str, Any] = dict(await get_attrs())
    result["definitions"] = filtered_attrdefs

    return result


def _check_net_data_changed(data: dict) -> bool:
    changed = False

    if system.net.has_wifi_support():
        wifi_config = system.net.get_wifi_config()
        old_wifi_config = data.get("wifi_config")
        if old_wifi_config != wifi_config:
            data["wifi_config"] = wifi_config
            changed = True

    if system.net.has_ip_support():
        ip_config = system.net.get_ip_config()
        old_ip_config = data.get("ip_config")
        if old_ip_config != ip_config:
            data["ip_config"] = ip_config
            changed = True

    return changed


async def _attrs_watch_loop() -> None:
    last_net_data = {}

    try:
        while True:
            changed = False
            try:
                if _check_net_data_changed(last_net_data):
                    logger.debug("network attributes data changed")
                    changed = True
            except Exception as e:
                logger.error("network attributes data check failed: %s", e, exc_info=True)

            if changed:
                await device_events.trigger_update()

            await asyncio.sleep(NETWORK_ATTRS_WATCH_INTERVAL)
    except asyncio.CancelledError:
        logger.debug("attributes watch task cancelled")


async def init() -> None:
    global _attrs_watch_task

    logger.debug("starting attributes watch task")
    _attrs_watch_task = asyncio.create_task(_attrs_watch_loop())


async def cleanup() -> None:
    logger.debug("stopping attributes watch task")
    if _attrs_watch_task:
        _attrs_watch_task.cancel()
        await _attrs_watch_task



from __future__ import annotations

import asyncio
import json
import logging
import time

from datetime import datetime

from tornado import httpclient

from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.typing import Attributes, GenericJSONDict
from qtoggleserver.slaves import utils as salves_utils
from qtoggleserver.system import ap, dhcp, dns, net
from qtoggleserver.utils import asyncio as asyncio_utils
from qtoggleserver.utils import cmd as cmd_utils

from .exceptions import DiscoverException


_PATH_PREFIXES = ["", "/api"]
_INTERFACE_CACHE_TIMEOUT = 60

logger = logging.getLogger(__name__)

_discover_task: asyncio.Task | None = None
_discovered_devices: dict[str, DiscoveredDevice] | None = None
_finish_timer: asyncio_utils.Timer | None = None

# TODO: replace these with a common cache service with integrated timeout management
_interface: tuple[str] | None = None
_interface_time: float = 0


class DiscoveredDevice:
    def __init__(
        self, ap_client: ap.APClient, scheme: str, port: int, path: str, admin_password: str, attrs: Attributes
    ) -> None:
        self.ap_client: ap.APClient = ap_client
        self.scheme: str = scheme
        self.port: int = port
        self.path: str = path
        self.admin_password: str = admin_password
        self.attrs: Attributes = attrs

    def __str__(self) -> str:
        return f"discovered AP device {self.attrs['name']} at {self.ap_client.ip_address}"

    def to_json(self) -> GenericJSONDict:
        return {
            "mac_address": self.ap_client.mac_address,
            "ip_address": self.ap_client.ip_address,
            "timestamp": datetime.timestamp(self.ap_client.moment),
            "scheme": self.scheme,
            "hostname": self.ap_client.hostname,
            "port": self.port,
            "path": self.path,
            "admin_password": self.admin_password,
            "attrs": self.attrs,
        }


def get_interface() -> str | None:
    global _interface
    global _interface_time

    ap_settings = settings.slaves.discover.ap
    if ap_settings.interface:
        return ap_settings.interface

    if ap_settings.interface_cmd:
        now = time.time()
        if now - _interface_time > _INTERFACE_CACHE_TIMEOUT:
            _interface = None

        if _interface is not None:
            return _interface[0]

        result = cmd_utils.run_get_cmd(
            ap_settings.interface_cmd,
            cmd_name="AP client discover interface",
            exc_class=DiscoverException,
            required_fields=["interface"],
        )

        _interface = (result.get("interface") or None,)
        _interface_time = now

        return _interface[0]
    else:
        return None


async def discover(timeout: int) -> None:
    global _discover_task
    global _discovered_devices
    global _finish_timer

    # If discover task is already running, just await it instead of starting a new one
    if _discover_task:
        await _discover_task
        return

    _discover_task = asyncio.create_task(_discover(timeout))

    if _finish_timer:
        _finish_timer.cancel()

    _finish_timer = asyncio_utils.Timer(settings.slaves.discover.ap.finish_timeout, _finish_timeout)

    devices_list = await _discover_task

    logger.debug("discovered %d devices", len(devices_list))

    _discovered_devices = {d.attrs["name"]: d for d in devices_list}


async def finish() -> None:
    global _discover_task
    global _discovered_devices
    global _finish_timer

    logger.debug("finishing")

    if _discover_task:
        _discover_task.cancel()
        try:
            await _discover_task
        except Exception as e:
            logger.error("discover task error: %s", e, exc_info=True)

        _discover_task = None

    if ap.is_running():
        await ap.stop()

    if _finish_timer:
        _finish_timer.cancel()
        _finish_timer = None

    _discovered_devices = None


def get_discovered_devices() -> dict[str, DiscoveredDevice] | None:
    return _discovered_devices


async def configure(discovered_device: DiscoveredDevice, attrs: Attributes) -> DiscoveredDevice:
    ap_client = discovered_device.ap_client

    # Wi-Fi settings of the real (target) network
    if "wifi_ssid" not in attrs:
        logger.warning("no SSID/PSK available to configure device")

    if "admin_password" not in attrs:
        logger.debug("generating password for %s", discovered_device)
        discovered_device.admin_password = salves_utils.generate_password(
            core_device_attrs.admin_password_hash, discovered_device.ap_client.mac_address, "admin"
        )

        attrs["admin_password"] = discovered_device.admin_password

        # Also set normal and view-only password, if exposed
        if "normal_password" in discovered_device.attrs:
            attrs["normal_password"] = salves_utils.generate_password(
                core_device_attrs.admin_password_hash, discovered_device.ap_client.mac_address, "normal"
            )

        if "viewonly_password" in discovered_device.attrs:
            attrs["viewonly_password"] = salves_utils.generate_password(
                core_device_attrs.admin_password_hash, discovered_device.ap_client.mac_address, "viewonly"
            )
    else:
        logger.debug("using supplied password for %s", discovered_device)
        discovered_device.admin_password = attrs["admin_password"]

    network_configured = attrs.get("wifi_ssid") is not None
    if network_configured:
        # Find client's future IP address first
        dhcp_interface = settings.slaves.discover.dhcp_interface or net.get_default_interface()
        if not dhcp_interface:
            raise DiscoverException("No DHCP interface")

        try:
            reply = await dhcp.request(
                interface=dhcp_interface,
                timeout=settings.slaves.discover.dhcp_timeout,
                mac_address=ap_client.mac_address,
                hostname=ap_client.hostname,
            )
        except dhcp.DHCPTimeout:
            logger.warning("could not determine future device IP address of %s", discovered_device)
            reply = None

        if reply:
            adjusted_ap_client = ap.APClient(
                mac_address=ap_client.mac_address,
                ip_address=reply.ip_address,
                hostname=ap_client.hostname,
                moment=ap_client.moment,
            )

            discovered_device = DiscoveredDevice(
                ap_client=adjusted_ap_client,
                scheme=discovered_device.scheme,
                port=discovered_device.port,
                path=discovered_device.path,
                admin_password=discovered_device.admin_password,
                attrs=discovered_device.attrs,
            )

            dns.set_custom_dns_mapping(ap_client.hostname, reply.ip_address, timeout=60)

    logger.debug("configuring %s", discovered_device)
    await ap_client.request("PATCH", f"{discovered_device.path}/device", body=attrs)
    logger.debug("%s successfully configured", discovered_device)

    # Remove configured device from discovered list
    _discovered_devices.pop(discovered_device.attrs["name"])

    if network_configured:
        logger.debug("waiting for %s to connect to new network", discovered_device)
        await asyncio.sleep(5)  # device requires at least 5 seconds to connect to new network
        start_time = time.time()
        while True:
            try:
                await discovered_device.ap_client.request(
                    "GET", "/device", no_log=True, admin_password=discovered_device.admin_password
                )
                logger.debug("%s connected to new network", discovered_device)
                break
            except Exception:
                if time.time() - start_time > settings.slaves.long_timeout:
                    logger.error("timeout waiting for %s to connect to new network", discovered_device)
                    break

                await asyncio.sleep(1)

    return discovered_device


async def _discover(timeout: int) -> list[DiscoveredDevice]:
    logger.debug("starting discovery")

    # (Re)start our AP
    ap_settings = settings.slaves.discover.ap

    if ap.is_running():
        await ap.stop()

    interface = get_interface()
    if not interface:
        raise DiscoverException("No AP interface available")

    ap.start(
        interface=interface,
        ssid=ap_settings.ssid,
        psk=ap_settings.psk,
        own_ip=ap_settings.own_ip,
        mask_len=ap_settings.mask_len,
        start_ip=ap_settings.start_ip,
        stop_ip=ap_settings.stop_ip,
        hostapd_binary=ap_settings.hostapd_binary,
        hostapd_cli_binary=ap_settings.hostapd_cli_binary,
        dnsmasq_binary=ap_settings.dnsmasq_binary,
        hostapd_log=ap_settings.hostapd_log,
        dnsmasq_log=ap_settings.dnsmasq_log,
    )

    # Wait for clients to connect to our AP
    try:
        for _ in range(timeout):
            await asyncio.sleep(1)
            await ap.update()
            if not ap.is_running():
                raise DiscoverException("AP stopped unexpectedly")

        clients = ap.get_clients()
    except asyncio.CancelledError:
        logger.debug("discover task cancelled")
        clients = []

    for client in clients:
        logger.debug(
            'discovered AP client "%s" with IP %s and MAC %s', client.hostname, client.ip_address, client.mac_address
        )

    if not clients:
        logger.debug("no clients discovered")

    # See if we're really dealing with qToggle devices
    discovered_devices = []
    for client in clients:
        try:
            discovered_devices.append(await _query_client(client))
        except Exception as e:
            logger.error("client query failed: %s", e, exc_info=True)

    return discovered_devices


async def _finish_timeout() -> None:
    global _finish_timer

    logger.debug("finish timeout")
    _finish_timer = None
    await finish()


async def _query_client(ap_client: ap.APClient) -> DiscoveredDevice | None:
    logger.debug("querying %s", ap_client)

    for prefix in _PATH_PREFIXES:
        try:
            attrs = await ap_client.request("GET", f"{prefix}/device")
            break
        except (httpclient.HTTPError, json.JSONDecodeError):
            continue
    else:
        raise DiscoverException("Could not find device API endpoint")

    if not isinstance(attrs, dict):
        logger.error("%s has returned invalid device attributes response", ap_client)
        return None

    for field in ["name", "vendor", "version"]:
        if field not in attrs:
            logger.error("%s has returned invalid device attributes response", ap_client)
            return None

    return DiscoveredDevice(
        ap_client=ap_client,
        scheme="http",  # TODO: add support for HTTPS schemes
        port=80,
        path=prefix,
        admin_password="",
        attrs=attrs,
    )


async def init() -> None:
    pass


async def cleanup() -> None:
    await finish()



import abc
import logging

from datetime import datetime

from jinja2 import Environment, Template

from qtoggleserver.conf import settings
from qtoggleserver.core import device
from qtoggleserver.core import events as core_events
from qtoggleserver.core import main as core_main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import Attribute, Attributes, NullablePortValue
from qtoggleserver.slaves import devices as slaves_devices

from .filtereventhandler import FilterEventHandler


logger = logging.getLogger(__name__)


class TemplateNotificationsHandler(FilterEventHandler, metaclass=abc.ABCMeta):
    DEFAULT_TEMPLATES = {  # TODO: i18n
        "value-change": {
            "title": "{{port.get_display_name()}} is {{port.get_display_value()}}",
            "body": (
                "Port {{port.get_display_name()}} was {{port.get_display_value(old_value)}} "
                "and is now {{port.get_display_value(new_value)}}."
            ),
        },
        "port-update": {
            "title": "{{port.get_display_name()}} has been updated",
            "body": "Port {{port.get_display_name()}} attributes have been updated.",
        },
        "port-add": {"title": "{{port.get_display_name()}} has been added", "body": None},
        "port-remove": {"title": "{{port.get_display_name()}} has been removed", "body": None},
        "device-update": {
            "title": "{{device.get_display_name()}} has been updated",
            "body": "Device {{device.get_display_name()}} attributes have been updated.",
        },
        "full-update": {
            "title": "{{device.get_display_name()}} has been updated",
            "body": "Device {{device.get_display_name()}} has been fully updated.",
        },
        "slave-device-update": {
            "title": "{{slave.get_display_name()}} has been updated",
            "body": "Device {{slave.get_display_name()}} attributes have been updated.",
        },
        "slave-device-add": {"title": "{{slave.get_display_name()}} has been added", "body": None},
        "slave-device-remove": {"title": "{{slave.get_display_name()}} has been removed", "body": None},
    }

    logger = logger

    def __init__(
        self,
        *,
        template: dict[str, str] | None = None,
        templates: dict[str, dict[str, str]] | None = None,
        skip_startup: int | bool = True,
        filter: dict | None = None,
        name: str | None = None,
    ) -> None:
        self._skip_startup: int | bool = skip_startup

        # "template" has the highest precedence; then comes "templates" and then comes "DEFAULT_TEMPLATES"
        if template is not None:
            templates = {k: template for k in self.DEFAULT_TEMPLATES.keys()}
        if templates is None:
            templates = self.DEFAULT_TEMPLATES

        # Convert template strings to jinja2 templates
        self._j2env: Environment = Environment(enable_async=True)
        self._templates: dict[str, None | Template | dict[str, Template]] = {}
        for type_, ts in templates.items():
            if isinstance(ts, dict):
                self._templates[type_] = {}
                for k, t in ts.items():
                    self._templates[type_][k] = self.make_template(t) if t is not None else None
            elif isinstance(ts, str):
                self._templates[type_] = self.make_template(ts)
            else:
                self._templates[type_] = None

        super().__init__(name=name, filter=filter)

    def get_j2env(self) -> Environment:
        return self._j2env

    def make_template(self, source: str) -> Template:
        return self.get_j2env().from_string(source)

    async def render(self, event_type: str, context: dict) -> None | str | dict[str, str]:
        template = self._templates[event_type]

        if isinstance(template, dict):
            return {k: await t.render_async(context) if t is not None else None for k, t in template.items()}
        elif isinstance(template, Template):
            return await template.render_async(context)
        else:
            return None

    def get_common_context(self, event: core_events.Event) -> dict:
        timestamp = event.get_timestamp()
        if timestamp:
            moment = datetime.fromtimestamp(timestamp)
        else:
            moment = None

        return {
            "event": event,
            "type": event.get_type(),
            "timestamp": timestamp,
            "moment": moment,
            "display_moment": moment.strftime("%c") if moment else "",
            "public_url": settings.public_url,
            "device_attrs": self.get_device_attrs(),
            "port_values": self.get_port_values(),
            "port_attrs": self.get_port_attrs(),
            "slave_attrs": self.get_slave_attrs(),
        }

    async def push_message(self, event: core_events.Event, title: str, body: str | None = None, **kwargs) -> None:
        pass

    async def push_template_message(self, event: core_events.Event, context: dict) -> None:
        rendered = await self.render(event.get_type(), context)
        if isinstance(rendered, dict):
            await self.push_message(event, **rendered)
        elif isinstance(rendered, str):
            await self.push_message(event, title=rendered)

    async def accepts(self, *args, **kwargs) -> bool:
        # Skip notifications during startup
        if self._skip_startup:
            if not core_main.is_ready():
                return False

            if isinstance(self._skip_startup, (int, float)) and core_main.uptime() < self._skip_startup:
                return False

        return await super().accepts(*args, **kwargs)

    async def on_value_change(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_value: NullablePortValue,
        new_value: NullablePortValue,
        attrs: Attributes,
    ) -> None:
        context = self.get_common_context(event)
        context.update({"port": port, "old_value": old_value, "new_value": new_value, "attrs": attrs})

        await self.push_template_message(event, context)

    async def on_port_update(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: dict[str, tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes,
    ) -> None:
        context = self.get_common_context(event)
        context.update(
            {
                "port": port,
                "old_attrs": old_attrs,
                "new_attrs": new_attrs,
                "changed_attrs": changed_attrs,
                "added_attrs": added_attrs,
                "removed_attrs": removed_attrs,
                "value": port.get_last_read_value(),
            }
        )

        await self.push_template_message(event, context)

    async def on_port_add(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        context = self.get_common_context(event)
        context.update({"port": port, "attrs": attrs, "value": port.get_last_read_value()})

        await self.push_template_message(event, context)

    async def on_port_remove(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        context = self.get_common_context(event)
        context.update({"port": port, "attrs": attrs, "value": port.get_last_read_value()})

        await self.push_template_message(event, context)

    async def on_device_update(
        self,
        event: core_events.Event,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: dict[str, tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes,
    ) -> None:
        context = self.get_common_context(event)
        context.update(
            {
                "device": device,
                "old_attrs": old_attrs,
                "new_attrs": new_attrs,
                "changed_attrs": changed_attrs,
                "added_attrs": added_attrs,
                "removed_attrs": removed_attrs,
            }
        )

        await self.push_template_message(event, context)

    async def on_full_update(self, event: core_events.Event) -> None:
        context = self.get_common_context(event)

        await self.push_template_message(event, context)

    async def on_slave_device_update(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: dict[str, tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes,
    ) -> None:
        context = self.get_common_context(event)
        context.update(
            {
                "slave": slave,
                "old_attrs": old_attrs,
                "new_attrs": new_attrs,
                "changed_attrs": changed_attrs,
                "added_attrs": added_attrs,
                "removed_attrs": removed_attrs,
            }
        )

        await self.push_template_message(event, context)

    async def on_slave_device_add(
        self, event: core_events.Event, slave: slaves_devices.Slave, attrs: Attributes
    ) -> None:
        context = self.get_common_context(event)
        context.update({"slave": slave, "attrs": attrs})

        await self.push_template_message(event, context)

    async def on_slave_device_remove(
        self, event: core_events.Event, slave: slaves_devices.Slave, attrs: Attributes
    ) -> None:
        context = self.get_common_context(event)
        context.update({"slave": slave, "attrs": attrs})

        await self.push_template_message(event, context)



import json
import math

from datetime import date, datetime
from typing import Any, cast

import jsonpointer


JSON_CONTENT_TYPE = "application/json; charset=utf-8"

TYPE_FIELD = "__t"
VALUE_FIELD = "__v"
DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
DATE_TYPE = "__d"
DATETIME_TYPE = "__dt"

DATETIME_FORMAT_ISO = "%Y-%m-%dT%H:%M:%SZ"
DATE_FORMAT_ISO = "%Y-%m-%d"
DATETIME_FORMAT_ISO_LEN = len(DATETIME_FORMAT_ISO)
DATE_FORMAT_ISO_LEN = len(DATE_FORMAT_ISO)

EXTRA_TYPES_NONE = ""
EXTRA_TYPES_ISO = "iso"
EXTRA_TYPES_EXTENDED = "extended"


def _replace_nan_inf_rec(obj: Any, replace_value: Any) -> Any:
    new_obj: Any
    if isinstance(obj, dict):
        new_obj = {}
        for k, v in obj.items():
            if isinstance(v, float) and (math.isnan(v) or math.isinf(v)):
                new_obj[k] = replace_value
            else:
                new_obj[k] = _replace_nan_inf_rec(v, replace_value)
    elif isinstance(obj, (list, tuple, set)):
        new_obj = []
        for e in obj:
            if isinstance(e, float) and (math.isnan(e) or math.isinf(e)):
                new_obj.append(replace_value)
            else:
                new_obj.append(_replace_nan_inf_rec(e, replace_value))
    else:
        new_obj = obj

    return new_obj


def _resolve_refs_rec(obj: Any, root_obj: Any) -> Any:
    if isinstance(obj, dict):
        if len(obj.keys()) == 1 and list(obj.keys())[0] == "$ref":
            ref = list(obj.values())[0]
            ref = ref[1:]  # skip starting hash
            return jsonpointer.resolve_pointer(root_obj, ref)

        for k, v in obj.items():
            obj[k] = _resolve_refs_rec(v, root_obj)
    elif isinstance(obj, list):
        for i, e in enumerate(obj):
            obj[i] = _resolve_refs_rec(e, root_obj)

    return obj


def encode_default_json_iso(obj: Any) -> Any:
    if isinstance(obj, datetime):
        return obj.strftime(DATETIME_FORMAT_ISO)
    elif isinstance(obj, date):
        return obj.strftime(DATE_FORMAT)
    elif isinstance(obj, (set, tuple)):
        return list(obj)
    else:
        raise TypeError()


def encode_default_json_extended(obj: Any) -> Any:
    if isinstance(obj, datetime):
        return {TYPE_FIELD: DATETIME_TYPE, VALUE_FIELD: obj.strftime(DATETIME_FORMAT)}
    elif isinstance(obj, date):
        return {TYPE_FIELD: DATE_TYPE, VALUE_FIELD: obj.strftime(DATE_FORMAT)}
    elif isinstance(obj, (set, tuple)):
        return list(obj)
    else:
        raise TypeError()


def decode_json_hook_iso(obj: dict) -> Any:
    for k, v in obj.items():
        if isinstance(v, str):
            if len(v) == DATETIME_FORMAT_ISO_LEN:
                try:
                    obj[k] = datetime.strptime(v, DATETIME_FORMAT_ISO)
                except ValueError:
                    pass
            elif len(v) == DATE_FORMAT_ISO_LEN:
                try:
                    obj[k] = datetime.strptime(v, DATE_FORMAT_ISO).date()
                except ValueError:
                    pass

    return obj


def decode_json_hook_extended(obj: dict[Any, Any]) -> Any:
    __t = obj.get(TYPE_FIELD)
    if __t is not None:
        __v = cast(str, obj.get(VALUE_FIELD))
        if __t == DATE_TYPE:
            try:
                return datetime.strptime(__v, DATE_FORMAT).date()
            except ValueError:
                pass
        elif __t == DATETIME_TYPE:
            try:
                return datetime.strptime(__v, DATETIME_FORMAT)
            except ValueError:
                pass

    return obj


def dumps(obj: Any, extra_types: str = EXTRA_TYPES_NONE, **kwargs) -> str:
    # Treat primitive types separately to gain just a bit of performance
    if isinstance(obj, str):
        return '"' + obj + '"'
    elif isinstance(obj, bool):
        return ["false", "true"][obj]
    elif isinstance(obj, (int, float)):
        if math.isinf(obj) or math.isnan(obj):
            return "null"

        return str(obj)
    elif obj is None:
        return "null"
    else:
        if extra_types == EXTRA_TYPES_EXTENDED:
            return json.dumps(obj, default=encode_default_json_extended, allow_nan=True, **kwargs)
        elif extra_types == EXTRA_TYPES_ISO:
            return json.dumps(obj, default=encode_default_json_iso, allow_nan=False, **kwargs)
        else:
            try:
                return json.dumps(obj, allow_nan=False, **kwargs)
            except ValueError:
                # Retry again by replacing Infinity and NaN values with None
                obj = _replace_nan_inf_rec(obj, replace_value=None)
                return json.dumps(obj, allow_nan=False, **kwargs)


def loads(s: str | bytes, resolve_refs: bool = False, extra_types: str = EXTRA_TYPES_NONE, **kwargs) -> Any:
    if extra_types == EXTRA_TYPES_EXTENDED:
        object_hook = decode_json_hook_extended
    elif extra_types == EXTRA_TYPES_ISO:
        object_hook = decode_json_hook_iso
    else:
        object_hook = None

    obj = json.loads(s, object_hook=object_hook, **kwargs)
    if resolve_refs:
        obj = _resolve_refs_rec(obj, root_obj=obj)

    return obj



import asyncio
import logging
import re
import subprocess
import tempfile
import time

from typing import TextIO, cast

import psutil

from .exceptions import APException


BINARY = "hostapd"
CLI_BINARY = "hostapd_cli"

CONF_TEMPLATE = (
    "ssid={ssid}\n"
    "wpa_passphrase={psk}\n"
    "wpa=2\n"
    "wpa_key_mgmt=WPA-PSK\n"
    "interface={interface}\n"
    "channel=1\n"
    "driver=nl80211\n"
    "ctrl_interface=/var/run/hostapd\n"
)

CONF_NO_PSK_TEMPLATE = (
    "ssid={ssid}\ninterface={interface}\nchannel=1\ndriver=nl80211\nctrl_interface=/var/run/hostapd\n"
)

STOP_TIMEOUT = 2


logger = logging.getLogger(__name__)


class HostAPDException(APException):
    pass


class HostAPD:
    def __init__(
        self,
        ssid: str,
        psk: str | None,
        interface: str,
        hostapd_binary: str | None = None,
        hostapd_cli_binary: str | None = None,
        hostapd_log: str | None = None,
    ) -> None:
        self._ssid: str = ssid
        self._psk: str | None = psk
        self._interface: str = interface
        self._binary: str | None = hostapd_binary
        self._cli_binary: str | None = hostapd_cli_binary
        self._log: str | None = hostapd_log

        self._conf_file: TextIO | None = None
        self._log_file: TextIO | None = None
        self._process: subprocess.Popen | None = None

    def is_alive(self) -> bool:
        return (self._process is not None) and (self._process.poll() is None)

    def is_running(self) -> bool:
        return self._process is not None

    def start(self) -> None:
        logger.debug(
            'starting hostapd with SSID "%s" and PSK %s', self._ssid, f'"{self._psk}"' if self._psk else "empty"
        )

        binary = self._binary or self._find_binary(BINARY)
        if not binary:
            raise HostAPDException("Could not find %s binary", BINARY)

        conf_template = CONF_TEMPLATE if self._psk else CONF_NO_PSK_TEMPLATE
        conf = conf_template.format(ssid=self._ssid, psk=self._psk, interface=self._interface)

        if self._log:
            self._log_file = cast(TextIO, open(self._log, "w"))
        self._conf_file = cast(TextIO, tempfile.NamedTemporaryFile(mode="wt"))
        self._conf_file.write(conf)
        self._conf_file.flush()

        self._process = subprocess.Popen(
            [binary, self._conf_file.name], stdout=self._log_file, stderr=subprocess.STDOUT
        )

    async def stop(self) -> None:
        logger.debug("stopping hostapd")

        if self._process:
            self._disassociate_clients()
            await asyncio.sleep(1)  # allow 1 second for hostapd to honor the command

            self._process.terminate()

            begin_time = time.time()
            while self._process.poll() is None:
                await asyncio.sleep(0.1)
                if time.time() - begin_time > STOP_TIMEOUT:
                    break

            # If process could not be stopped in time, kill it
            if self._process.poll() is None:
                logger.error("failed to stop hostapd within %d seconds, killing it", STOP_TIMEOUT)
                self._process.kill()
                await asyncio.sleep(1)
                self._process.poll()  # we want no zombies

            self._process = None

        if self._conf_file:
            self._conf_file.close()
            self._conf_file = None

        if self._log_file:
            self._log_file.close()
            self._log_file = None

    def _disassociate_clients(self) -> None:
        cli_binary = self._cli_binary or self._find_binary(CLI_BINARY)
        if not cli_binary:
            raise HostAPDException("Could not find %s binary", CLI_BINARY)

        try:
            output = subprocess.check_output([cli_binary, "list"], stderr=subprocess.DEVNULL).decode().strip()
        except subprocess.CalledProcessError:
            logger.error("command hostapd_cli list failed")
            return

        lines = output.split("\n")
        mac_addresses = [line for line in lines if re.match(r"^[a-f0-9:]{17}$", line)]

        for mac_address in mac_addresses:
            logger.debug("disassociating client with MAC %s", mac_address.upper())
            try:
                subprocess.check_call(
                    [cli_binary, "disassociate", mac_address], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
                )
            except subprocess.CalledProcessError:
                logger.error("command hostapd_cli disassociate failed")

    def _find_binary(self, binary: str) -> str | None:
        try:
            return subprocess.check_output(["which", binary], stderr=subprocess.DEVNULL).decode().strip()
        except subprocess.CalledProcessError:
            return None

    @staticmethod
    def is_already_running() -> bool:
        return BINARY in (p.name() for p in psutil.process_iter())



from .base import EvalContext, EvalResult
from .functions import Function, function


@function("IF")
class IfFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    async def _eval(self, context: EvalContext) -> EvalResult:
        condition_value = await self.args[0].eval(context)
        if condition_value:
            return await self.args[1].eval(context)
        else:
            return await self.args[2].eval(context)


@function("EQ")
class EqFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] == eval_args[1])


@function("GT")
class GTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] > eval_args[1])


@function("GTE")
class GTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] >= eval_args[1])


@function("LT")
class LTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] < eval_args[1])


@function("LTE")
class LTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] <= eval_args[1])



import logging

from qtoggleserver import persist

from . import devices as slaves_devices
from . import discover
from . import ports as slaves_ports


logger = logging.getLogger(__name__)


async def reset_ports() -> None:
    logger.debug("clearing slave ports persisted data")
    await persist.remove(slaves_ports.SlavePort.PERSIST_COLLECTION)


async def reset_slaves() -> None:
    logger.debug("clearing slaves persisted data")
    await persist.remove("slaves")


async def init() -> None:
    logger.debug("loading devices")
    await slaves_devices.load()

    logger.debug("initializing discover mechanism")
    await discover.init()


async def cleanup() -> None:
    logger.debug("cleaning up discover mechanism")
    await discover.cleanup()

    logger.debug("cleaning up devices")
    await slaves_devices.cleanup()



import logging

import qui

from qtoggleserver import system, version
from qtoggleserver.conf import settings

from .base import APIHandler, BaseHandler


__all__ = ["APIHandler", "BaseHandler"]


logger = logging.getLogger(__name__)


async def init() -> None:
    from qtoggleserver.core import history
    from qtoggleserver.slaves.discover import is_enabled as is_discover_enabled

    if settings.frontend.enabled:
        logger.debug("initializing QUI")

        qui.configure(
            name="qtoggleserver",
            display_name=settings.frontend.display_name,
            display_short_name=settings.frontend.display_short_name,
            description=settings.frontend.description,
            version=version.VERSION,
            debug=settings.frontend.debug,
            static_url=settings.frontend.static_url,
            extra_context=dict(
                slaves_enabled=settings.slaves.enabled,
                discover_enabled=is_discover_enabled(),
                history_enabled=history.is_enabled(),
                setup_mode=system.is_setup_mode(),
            ),
        )


async def cleanup() -> None:
    pass



from .base import Event, Handler
from .device import DeviceEvent, DeviceUpdate, FullUpdate
from .handlers import cleanup as cleanup_handlers
from .handlers import disable, enable, register_handler, trigger
from .handlers import init as init_handlers
from .port import PortAdd, PortEvent, PortRemove, PortUpdate, ValueChange


__all__ = [
    "DeviceEvent",
    "DeviceUpdate",
    "Event",
    "FullUpdate",
    "Handler",
    "PortAdd",
    "PortEvent",
    "PortRemove",
    "PortUpdate",
    "ValueChange",
    "disable",
    "enable",
    "register_handler",
]


async def trigger_full_update() -> None:
    await trigger(FullUpdate())


async def init() -> None:
    await init_handlers()


async def cleanup() -> None:
    await cleanup_handlers()



import logging

from qtoggleserver import persist
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList, NullablePortValue, PortValueChoices


logger = logging.getLogger(__name__)

_vport_args: dict[str, GenericJSONDict] = {}


class VirtualPort(core_ports.Port):
    WRITABLE = True
    VIRTUAL = True

    def __init__(
        self,
        id_: str,
        type_: str,
        min_: float | None,
        max_: float | None,
        integer: bool | None,
        step: float | None,
        choices: PortValueChoices | None,
    ) -> None:
        super().__init__(id_)

        self._type: str = type_
        self._min: float | None = min_
        self._max: float | None = max_
        self._integer: bool | None = integer
        self._step: float | None = step
        self._choices: PortValueChoices | None = choices
        self._virtual_value: NullablePortValue = None

        self._value = None

    def map_id(self, new_id: str) -> None:
        raise core_ports.PortError("Virtual ports cannot be mapped")

    async def read_value(self) -> NullablePortValue:
        return self._virtual_value

    async def write_value(self, value: NullablePortValue) -> None:
        self._virtual_value = value


async def add(
    id_: str,
    type_: str,
    min_: float | None,
    max_: float | None,
    integer: bool | None,
    step: float | None,
    choices: PortValueChoices | None,
) -> None:
    settings = {
        "id": id_,
        "type": type_,
        "min": min_,
        "max": max_,
        "integer": integer,
        "step": step,
        "choices": choices,
    }

    _vport_args[id_] = settings

    logger.debug("saving virtual port settings for %s", id_)
    await persist.replace("vports", id_, settings)


async def remove(port_id: str) -> None:
    _vport_args.pop(port_id, None)
    logger.debug("removing virtual port settings for %s", port_id)
    await persist.remove("vports", filt={"id": port_id})


def all_port_args() -> GenericJSONList:
    return [{"driver": VirtualPort, "id_": port_id, **args} for port_id, args in _vport_args.items()]


async def init() -> None:
    for entry in await persist.query("vports"):
        _vport_args[entry["id"]] = {
            "type_": entry.get("type") or core_ports.TYPE_NUMBER,
            "min_": entry.get("min"),
            "max_": entry.get("max"),
            "integer": entry.get("integer"),
            "step": entry.get("step"),
            "choices": entry.get("choices"),
        }

        logger.debug("loaded virtual port settings for %s", entry["id"])

    await core_ports.load(all_port_args())


async def reset() -> None:
    logger.debug("clearing virtual ports persisted data")
    await persist.remove("vports")



import importlib
import logging

from qtoggleserver import persist
from qtoggleserver.utils import json as json_utils

from . import attrs as device_attrs


_PERSISTED_MODULE_ATTRS = ["admin_password_hash", "normal_password_hash", "viewonly_password_hash"]

logger = logging.getLogger(__name__)


def get_display_name() -> str:
    # Used by template notifications
    return device_attrs.display_name or device_attrs.name


async def load() -> None:
    data = await persist.get_value("device", default={})

    # Load persisted module attributes
    for name in _PERSISTED_MODULE_ATTRS:
        value = data.get(name)
        if value is None:
            continue

        # A few attributes may carry sensitive information, so treat them separately and do not log their values
        if name.count("password") or name == "wifi_key":
            logger.debug("loaded %s", name)
        else:
            logger.debug("loaded %s = %s", name, json_utils.dumps(value))

        setattr(device_attrs, name, value)

    # Load persisted device attributes
    attrdefs = device_attrs.get_attrdefs()
    attrs = {}
    for name, attrdef in attrdefs.items():
        if attrdef.get("persisted") and data.get(name) is not None:
            attrs[name] = data[name]
    if attrs:
        await device_attrs.set_attrs(attrs)

    # Hash empty passwords
    if not device_attrs.admin_password_hash:
        device_attrs.admin_password_hash = device_attrs.EMPTY_PASSWORD_HASH
    if not device_attrs.normal_password_hash:
        device_attrs.normal_password_hash = device_attrs.EMPTY_PASSWORD_HASH
    if not device_attrs.viewonly_password_hash:
        device_attrs.viewonly_password_hash = device_attrs.EMPTY_PASSWORD_HASH


async def save() -> None:
    data = {name: getattr(device_attrs, name) for name in _PERSISTED_MODULE_ATTRS}

    attrdefs = device_attrs.get_attrdefs()
    attrs = await device_attrs.get_attrs()
    for name, attrdef in attrdefs.items():
        if attrdef.get("persisted") and attrs.get(name) is not None:
            data[name] = attrs[name]

    logger.debug("saving persisted data")
    await persist.set_value("device", data)


async def reset(preserve_attrs: list[str] | None = None) -> None:
    preserve_attrs = preserve_attrs or []

    preserved_attrs = {}
    for name in preserve_attrs:
        preserved_attrs[name] = getattr(device_attrs, name, None)

    logger.debug("clearing persisted data")
    await persist.remove("device")
    importlib.reload(device_attrs)  # reloads device attributes to default values

    for name, value in preserved_attrs.items():
        setattr(device_attrs, name, value)


async def init() -> None:
    logger.debug("loading persisted data")
    await load()

    logger.debug("initializing attributes")
    await device_attrs.init()


async def cleanup() -> None:
    logger.debug("cleaning up attributes")
    await device_attrs.cleanup()



from .base import EvalContext, EvalResult
from .functions import Function, function


@function("BITAND")
class BitAndFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        r = -1
        for e in await self.eval_args(context):
            r &= int(e)

        return r


@function("BITOR")
class BitOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        r = 0
        for e in await self.eval_args(context):
            r |= int(e)

        return r


@function("BITNOT")
class BitNotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def _eval(self, context: EvalContext) -> EvalResult:
        return ~int((await self.eval_args(context))[0])


@function("BITXOR")
class BitXOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0]) ^ int(eval_args[1])


@function("SHL")
class SHLFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0]) << int(eval_args[1])


@function("SHR")
class SHRFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0]) >> int(eval_args[1])



import abc
import calendar

from datetime import datetime, time, timedelta, timezone

from qtoggleserver import system

from .base import EvalContext, EvalResult
from .exceptions import EvalSkipped, InvalidArgumentValue
from .functions import Function, function


class DateUnitFunction(Function, metaclass=abc.ABCMeta):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        if len(self.args) > 0:
            timestamp = int(await self.args[0].eval(context))
        else:
            timestamp = context.timestamp

        return self.extract_unit(datetime.fromtimestamp(timestamp))

    @abc.abstractmethod
    def extract_unit(self, dt: datetime) -> int:
        raise NotImplementedError()


@function("YEAR")
class YearFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.year


@function("MONTH")
class MonthFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.month


@function("DAY")
class DayFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.day


@function("DOW")
class DOWFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.weekday()


@function("LDOM")
class LDOMFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return calendar.monthrange(dt.year, dt.month)[1]


@function("HOUR")
class HourFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.hour


@function("MINUTE")
class MinuteFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.minute


@function("SECOND")
class SecondFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.second


@function("MILLISECOND")
class MillisecondFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = {"asap"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return int(context.now_ms % 1000)


@function("MINUTEDAY")
class MinuteDayFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.hour * 60 + dt.minute


@function("SECONDDAY")
class SecondDayFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime) -> int:
        return dt.hour * 3600 + dt.minute * 60 + dt.second


@function("DATE")
class DateFunction(Function):
    MIN_ARGS = MAX_ARGS = 6
    DEPS = {"second"}
    UNIT_INDEX = {u: i + 1 for i, u in enumerate(("year", "month", "day", "hour", "minute", "second"))}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        eval_args = [int(await self.args[i].eval(context)) for i in range(self.MIN_ARGS)]

        try:
            return int(datetime(*eval_args).timestamp())
        except ValueError as e:
            unit = str(e).split()[0]
            index = self.UNIT_INDEX.get(unit)
            if index is None:
                raise

            raise InvalidArgumentValue(index, eval_args[index])


@function("BOY")
class BOYFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.fromtimestamp(context.timestamp)

        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval(context))

        dt = datetime(now.year + n, 1, 1, 0, 0, 0)
        dt = dt.astimezone(tz=timezone.utc)

        return dt.timestamp()


@function("BOM")
class BOMFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.fromtimestamp(context.timestamp)
        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval(context))

        year, month = now.year, now.month
        if n >= 0:
            for _ in range(n):
                if month < 12:
                    month += 1
                else:
                    year += 1
                    month = 1
        else:
            for _ in range(-n):
                if month > 1:
                    month -= 1
                else:
                    year -= 1
                    month = 12

        dt = datetime(year, month, 1, 0, 0, 0)
        dt = dt.astimezone(tz=timezone.utc)

        return dt.timestamp()


@function("BOW")
class BOWFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 2
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        n = 0
        s = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval(context))
            if len(self.args) > 1:
                s = int(await self.args[1].eval(context))

        now = datetime.fromtimestamp(context.timestamp)
        dt = now.replace(hour=12)  # using midday practically avoids problems due to DST
        if s > 0:
            dt -= timedelta(days=dt.weekday() + 7 - s)
        else:
            dt -= timedelta(days=dt.weekday())

        year, month, day = dt.year, dt.month, dt.day
        if n >= 0:
            for _ in range(n):
                last_day = calendar.monthrange(year, month)[1]
                if day + 7 <= last_day:
                    day += 7
                else:
                    day = 7 - last_day + day
                    if month < 12:
                        month += 1
                    else:
                        year += 1
                        month = 1
        else:
            for _ in range(-n):
                if day > 7:
                    day -= 7
                else:
                    if month > 1:
                        month -= 1
                    else:
                        year -= 1
                        month = 12

                    last_day = calendar.monthrange(year, month)[1]
                    day = last_day - 7 + day

        dt = datetime(year, month, day)
        dt = dt.astimezone(tz=timezone.utc)

        return dt.timestamp()


@function("BOD")
class BODFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.fromtimestamp(context.timestamp)
        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval(context))
        dt = now + timedelta(days=n)
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)
        dt = dt.astimezone(tz=timezone.utc)

        return dt.timestamp()


@function("HMSINTERVAL")
class HMSIntervalFunction(Function):
    MIN_ARGS = MAX_ARGS = 6
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.fromtimestamp(context.timestamp).replace(microsecond=0)

        start_h, start_m, start_s, stop_h, stop_m, stop_s = await self.eval_args(context)

        if not (0 <= start_h <= 23):
            raise InvalidArgumentValue(1, start_h)

        if not (0 <= start_m <= 59):
            raise InvalidArgumentValue(2, start_m)

        if not (0 <= start_s <= 59):
            raise InvalidArgumentValue(3, start_s)

        if not (0 <= stop_h <= 23):
            raise InvalidArgumentValue(4, stop_h)

        if not (0 <= stop_m <= 59):
            raise InvalidArgumentValue(5, stop_m)

        if not (0 <= stop_s <= 59):
            raise InvalidArgumentValue(6, stop_s)

        start_time = time(int(start_h), int(start_m), int(start_s))
        stop_time = time(int(stop_h), int(stop_m), int(stop_s))

        start_dt = datetime.combine(now.date(), start_time)
        stop_dt = datetime.combine(now.date(), stop_time)

        return int(start_dt <= now <= stop_dt)


@function("MDINTERVAL")
class MDIntervalFunction(Function):
    MIN_ARGS = MAX_ARGS = 4
    DEPS = {"second"}

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.fromtimestamp(context.timestamp).replace(microsecond=0)
        start_m, start_d, stop_m, stop_d = await self.eval_args(context)

        if not (1 <= start_m <= 12):
            raise InvalidArgumentValue(1, start_m)

        try:
            start_dt = now.replace(month=int(start_m), day=int(start_d))
        except ValueError:
            raise InvalidArgumentValue(2, start_d)

        if not (1 <= stop_m <= 12):
            raise InvalidArgumentValue(3, stop_m)

        try:
            stop_dt = now.replace(month=int(stop_m), day=int(stop_d))
        except ValueError:
            raise InvalidArgumentValue(4, stop_d)

        return int(start_dt <= now <= stop_dt)



from __future__ import annotations

import abc
import logging

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils
from qtoggleserver.utils import dynload as dynload_utils


STATUS_IDLE = "idle"
STATUS_CHECKING = "checking"
STATUS_DOWNLOADING = "downloading"
STATUS_EXTRACTING = "extracting"
STATUS_VALIDATING = "validating"
STATUS_FLASHING = "flashing"
STATUS_RESTARTING = "restarting"
STATUS_ERROR = "error"


logger = logging.getLogger(__name__)

_driver: BaseDriver | None = None


class FWUpdateException(Exception):
    pass


class FWUpdateDisabled(FWUpdateException):
    pass


class BaseDriver(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    async def get_current_version(self) -> str:
        pass

    @abc.abstractmethod
    async def get_latest(self) -> tuple[str, str, str]:
        """Returns a tuple of (version, date, url)."""
        pass

    @abc.abstractmethod
    async def get_status(self) -> str:
        return STATUS_IDLE

    @abc.abstractmethod
    async def update_to_version(self, version: str) -> None:
        pass

    @abc.abstractmethod
    async def update_to_url(self, url: str) -> None:
        pass

    @abc.abstractmethod
    async def is_auto_update_enabled(self) -> bool:
        pass

    @abc.abstractmethod
    async def set_auto_update_enabled(self, enabled: bool) -> None:
        pass


def _get_fwupdate() -> BaseDriver:
    global _driver

    if not settings.system.fwupdate.driver:
        raise FWUpdateDisabled()

    if _driver is None:
        logger.debug("initializing fwupdate")

        driver_args = conf_utils.obj_to_dict(settings.system.fwupdate)
        driver_class_path = driver_args.pop("driver")

        driver_class = dynload_utils.load_attr(driver_class_path)
        _driver = driver_class(**driver_args)

    return _driver


async def get_current_version() -> str:
    current_version = await _get_fwupdate().get_current_version()
    logger.debug("current version: %s", current_version)

    return current_version


async def get_latest() -> tuple[str, str, str]:
    latest_version, latest_date, latest_url = await _get_fwupdate().get_latest()
    logger.debug("latest version: %s/%s at %s", latest_version, latest_date, latest_url)

    return latest_version, latest_date, latest_url


async def get_status() -> str:
    status = await _get_fwupdate().get_status()
    logger.debug("status: %s", status)

    return status


async def update_to_version(version: str) -> None:
    logger.debug("updating to version %s", version)

    await _get_fwupdate().update_to_version(version)


async def update_to_url(url: str) -> None:
    logger.debug("updating to url %s", url)

    await _get_fwupdate().update_to_url(url)


async def is_auto_update_enabled() -> bool:
    return await _get_fwupdate().is_auto_update_enabled()


async def set_auto_update_enabled(enabled: bool) -> None:
    logger.debug("auto-update %s", ["disabled", "enabled"][enabled])

    await _get_fwupdate().set_auto_update_enabled(enabled)



from __future__ import annotations

import abc
import asyncio
import functools
import hashlib
import logging

from collections.abc import Callable
from typing import Any, cast

from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.utils import asyncio as asyncio_utils
from qtoggleserver.utils import logging as logging_utils

from .exceptions import NotOurPort


logger = logging.getLogger(__package__)


class Peripheral(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
    RUNNER_CLASS = asyncio_utils.ThreadedRunner
    RUNNER_QUEUE_SIZE = 64

    logger = logger

    def __init__(
        self,
        *,
        params: dict[str, Any],
        name: str | None = None,
        id: str | None = None,
        static: bool = False,
        **kwargs,
    ) -> None:
        sorted_params = self._sorted_tuples_dict(params)
        auto_id_to_hash = f"{self.__class__.__module__}.{self.__class__.__name__}:{name}:{sorted_params}"
        auto_id = f"peripheral_{hashlib.sha256(auto_id_to_hash.encode()).hexdigest()[:8]}"

        self._params: dict[str, Any] = params
        self._name: str | None = name
        self._id: str = name or id or auto_id  # name will always be used as id, if supplied
        self._static: bool = static

        self._ports_by_id: dict[str, PeripheralPort] = {}
        self._enabled: bool = False
        self._online: bool = False
        self._runner: asyncio_utils.ThreadedRunner | None = None
        self._port_update_task: asyncio.Task | None = None

        logging_utils.LoggableMixin.__init__(self, self.get_id(), self.logger)

    @staticmethod
    def _sorted_tuples_dict(params: dict[str, Any]) -> tuple:
        def dict_reorder(d: dict) -> tuple:
            return tuple((k, dict_reorder(v)) if isinstance(v, dict) else v for k, v in sorted(d.items()))

        return dict_reorder(params)

    def __str__(self) -> str:
        return f"peripheral {self.get_id()}"

    def get_id(self) -> str:
        return self._id

    def get_name(self) -> str | None:
        return self._name

    def get_params(self) -> dict[str, Any]:
        return self._params

    def is_static(self) -> bool:
        return self._static

    def to_json(self) -> GenericJSONDict:
        return dict(self._params, id=self.get_id(), static=self.is_static(), name=self.get_name())

    async def get_port_args(self) -> list[dict[str, Any]]:
        port_args = await self.make_port_args()

        # Transform port classes to dicts with drivers
        transformed_port_args: list[dict[str, Any]] = [
            {"driver": pa} if isinstance(pa, type) else pa for pa in port_args
        ]

        # Supply the peripheral argument
        for pa in transformed_port_args:
            pa.setdefault("peripheral", self)

        return transformed_port_args

    @abc.abstractmethod
    async def make_port_args(self) -> list[dict[str, Any] | type[PeripheralPort]]:
        raise NotImplementedError()

    async def init_ports(self) -> None:
        port_args = await self.get_port_args()
        loaded_ports = await core_ports.load(port_args)
        self._ports_by_id = {cast(PeripheralPort, p).get_initial_id(): p for p in loaded_ports}

        # If no port has been loaded, there's no way for the user to enable the peripheral (via ports) so we have to
        # enable it manually, here.
        if not loaded_ports:
            await self.enable()

    async def cleanup_ports(self, persisted_data: bool) -> None:
        tasks = [asyncio.create_task(port.remove(persisted_data=persisted_data)) for port in self._ports_by_id.values()]
        if tasks:
            await asyncio.wait(tasks)

    async def add_port(self, port_args: dict[str, Any]) -> PeripheralPort:
        # Supply the peripheral argument
        port_args = dict(port_args)
        port_args.setdefault("peripheral", self)

        port = cast(PeripheralPort, (await core_ports.load([port_args]))[0])
        self._ports_by_id[port.get_initial_id()] = port
        return port

    async def remove_port(self, port_id: str, persisted_data: bool = False) -> None:
        if self._name and port_id.startswith(f"{self._name}."):
            port_id = port_id[len(self._name) + 1 :]

        try:
            port = self._ports_by_id.pop(port_id)
        except KeyError:
            raise NotOurPort(f"Port {port_id} does not belong to {self}") from None

        await port.remove(persisted_data=persisted_data)

    def get_ports(self) -> list[PeripheralPort]:
        return list(self._ports_by_id.values())

    def get_port(self, port_id: str) -> PeripheralPort | None:
        return self._ports_by_id.get(port_id)

    def is_enabled(self) -> bool:
        return self._enabled

    async def enable(self) -> None:
        if self._enabled:
            return

        self._enabled = True
        await self.handle_enable()
        self.debug("peripheral enabled")

    async def disable(self) -> None:
        if not self._enabled:
            return

        self._enabled = False
        await self.handle_disable()
        self.debug("peripheral disabled")

    async def check_disabled(self, exclude_port: PeripheralPort | None = None) -> None:
        if not self._enabled:
            return

        for port in self._ports_by_id.values():
            if port.is_enabled() and port != exclude_port:
                break
        else:
            self.debug("all ports are disabled, disabling peripheral")
            await self.disable()

    def is_online(self) -> bool:
        return self._enabled and self._online

    def set_online(self, online: bool) -> None:
        if online and not self._online:
            self.debug("is online")
            try:
                self.handle_online()
            except Exception:
                self.error("handle_online failed", exc_info=True)
        elif not online and self._online:
            self.debug("is offline")
            try:
                self.handle_offline()
            except Exception:
                self.error("handle_offline failed", exc_info=True)

        self._online = online

    def handle_offline(self) -> None:
        self.trigger_port_update_fire_and_forget()

    def handle_online(self) -> None:
        self.trigger_port_update_fire_and_forget()

    async def trigger_port_update(self, save: bool = False) -> None:
        self._port_update_task = None
        for port in self._ports_by_id.values():
            port.invalidate_attrs()
            if port.is_enabled():
                await port.trigger_update()
                if save:
                    port.save_asap()

    def trigger_port_update_fire_and_forget(self, save: bool = False) -> None:
        if self._port_update_task:
            return  # already scheduled

        self._port_update_task = asyncio.create_task(self.trigger_port_update(save))

    def get_runner(self) -> asyncio_utils.ThreadedRunner:
        if self._runner is None:
            self._runner = self.make_runner()

        return self._runner

    def make_runner(self) -> asyncio_utils.ThreadedRunner:
        self.debug("starting threaded runner")
        runner = self.RUNNER_CLASS(queue_size=self.RUNNER_QUEUE_SIZE)
        runner.start()

        return runner

    async def run_threaded(self, func: Callable, *args, **kwargs) -> Any:
        future = asyncio.get_running_loop().create_future()

        def callback(result: Any, exception: Exception | None) -> None:
            if future.cancelled():
                return

            if exception:
                future.set_exception(exception)
            else:
                future.set_result(result)

        runner = self.get_runner()
        runner.schedule_func(functools.partial(func, *args, **kwargs), callback)

        return await future

    async def handle_enable(self) -> None:
        pass

    async def handle_disable(self) -> None:
        pass

    async def handle_init(self) -> None:
        pass

    async def handle_cleanup(self) -> None:
        if self._port_update_task:
            await self._port_update_task

        if self._runner:
            self.debug("stopping threaded runner")
            await self._runner.stop()
            self.debug("threaded runner stopped")


# This needs to be imported here to avoid circular import issues
from .peripheralport import PeripheralPort  # noqa: E402



from typing import Any as _Any


source: str | None = None  # full path to the configuration file, automatically set at startup

debug: bool = False

public_url: str | None = None


logging: dict[str, _Any] = {
    "version": 1,
    "memory_logs_buffer_len": 10000,
    "formatters": {
        "default": {"format": "%(asctime)s: %(levelname)7s: [%(name)s] %(message)s", "datefmt": "%Y-%m-%d %H:%M:%S"}
    },
    "handlers": {"console": {"class": "logging.StreamHandler", "formatter": "default"}},
    "loggers": {
        # Double quotes are necessary to avoid HOCON key split
        '"asyncio"': {"level": "INFO"},
        '"bleak"': {"level": "INFO"},
        '"qtoggleserver.core.sessions"': {"level": "INFO"},
        '"qtoggleserver.persist"': {"level": "INFO"},
        '"qtoggleserver.drivers.persist"': {"level": "INFO"},
        '"qtoggleserver.utils.cmd"': {"level": "INFO"},
    },
    "root": {"level": "DEBUG", "handlers": ["console"]},
}


class core:
    class device_name:
        get_cmd: str | None = None
        set_cmd: str | None = None

    class passwords:
        set_cmd: str | None = None

    tick_interval: int = 50
    persist_interval: int = 2000
    event_queue_size: int = 1024
    max_client_time_skew: int = 300
    backup_support: bool = True
    history_support: bool = True
    history_janitor_interval: int = 3600
    listen_support: bool = True
    sequences_support: bool = True
    tls_support: bool = True
    virtual_ports: int = 1024


class server:
    addr: str = "0.0.0.0"
    port: int = 8888
    compress_response: bool = True

    class https:
        cert_file: str | None = None
        key_file: str | None = None


class persist:
    driver: str = "qtoggleserver.drivers.persist.JSONDriver"
    file_path: str = "qtoggleserver-data.json"


class system:
    setup_mode_cmd: str | None = None

    class date:
        set_cmd: str | None = None
        set_format: str | None = "%Y-%m-%dT%H:%M:%SZ"

    class timezone:
        get_cmd: str | None = None
        set_cmd: str | None = None

    class net:
        class wifi:
            get_cmd: str | None = None
            set_cmd: str | None = None

        class ip:
            get_cmd: str | None = None
            set_cmd: str | None = None

    class storage:
        path: str | None = None

    class temperature:
        get_cmd: str | None = None
        sensor_name: str | None = None
        sensor_index: int = 0
        min: int | None = 0
        max: int | None = 100

    class battery:
        get_cmd: str | None = None

    class fwupdate:
        driver: str | None = None


class frontend:
    enabled: bool = True
    debug: bool = False
    static_url: str | None = None
    display_name: str = "qToggleServer"
    display_short_name: str = "qToggleServer"
    description: str = "An application to control qToggleServer"  # TODO: i18n


class slaves:
    enabled: bool = True
    timeout: int = 10
    long_timeout: int = 60
    keepalive: int = 10
    retry_interval: int = 5
    retry_count: int = 3

    class discover:
        request_timeout: int = 5
        dhcp_timeout: int = 10
        dhcp_interface: str | None = None

        class ap:
            interface: str | None = None
            interface_cmd: str | None = None
            ssid: str = "qToggleSetup"
            psk: str | None = None
            own_ip: str = "192.168.43.1"
            mask_len: int = 24
            start_ip: str = "192.168.43.50"
            stop_ip: str = "192.168.43.250"
            hostapd_binary: str | None = None
            hostapd_cli_binary: str | None = None
            dnsmasq_binary: str | None = None
            hostapd_log: str = "/tmp/hostapd.log"
            dnsmasq_log: str = "/tmp/dnsmasq.log"
            finish_timeout: int = 300


class webhooks:
    enabled: bool = False


class reverse:
    enabled: bool = False
    retry_interval: int = 5


event_handlers: list[dict[str, _Any]] = []

peripherals: list[dict[str, _Any]] = []

ports: list[dict[str, _Any]] = []

port_mappings: dict[str, str] = {}



import asyncio

from collections.abc import Awaitable, Callable

from qtoggleserver import startup


def execute(main_func: Callable[[], Awaitable[bool]]) -> None:
    loop = asyncio.new_event_loop()

    loop.run_until_complete(startup.init_loop())
    loop.run_until_complete(startup.init())

    try:
        run_loop = loop.run_until_complete(main_func())
        if run_loop:
            loop.run_forever()
        loop.run_until_complete(startup.cleanup())

    finally:
        try:
            loop.run_until_complete(startup.cleanup_loop())
        except asyncio.CancelledError:
            pass  # ignore any cancelled errors

        loop.close()

    startup.logger.info("bye!")



from __future__ import annotations

import asyncio
import logging
import time

from collections.abc import Iterable

from qtoggleserver import persist, system
from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, PortValue
from qtoggleserver.utils import json as json_utils


_PERSIST_COLLECTION = "value_history"
_CACHE_TIMESTAMP_MIN_AGE = 3600 * 1000  # don't cache samples newer than this number of milliseconds ago

logger = logging.getLogger(__name__)

_history_event_handler: HistoryEventHandler | None = None
_sampling_task: asyncio.Task | None = None
_janitor_task: asyncio.Task | None = None

# Samples cached by port_id and timestamp
_samples_cache: dict[str, dict[int, PortValue]] = {}

# Used to schedule sample removal with remove_samples(..., background=True)
_pending_remove_samples: list[tuple[core_ports.BasePort, int | None, int | None]] = []


class HistoryEventHandler(core_events.Handler):
    FIRE_AND_FORGET = True

    async def handle_event(self, event: core_events.Event) -> None:
        if not isinstance(event, core_events.ValueChange):
            return  # we're only interested in port value changes

        if not system.date.has_real_date_time():
            return  # don't record history unless we've got real date/time

        port = event.get_port()
        history_interval = await port.get_history_interval()
        if history_interval != -1:
            return  # only consider ports with history interval set to special -1 (on value change)

        now_ms = int(time.time() * 1000)

        await save_sample(port, now_ms)
        port.set_history_last_timestamp(now_ms)


async def sampling_task() -> None:
    while True:
        try:
            await asyncio.sleep(1)

            if not system.date.has_real_date_time():
                continue  # don't record history unless we've got real date/time

            now_ms = int(time.time() * 1000)
            for port in core_ports.get_all():
                if not port.is_enabled():
                    continue

                history_last_timestamp = port.get_history_last_timestamp()
                history_interval = await port.get_history_interval()
                if history_interval <= 0:  # disabled or on value change
                    continue

                if now_ms - history_last_timestamp < history_interval * 1000:
                    continue

                await save_sample(port, now_ms)
                port.set_history_last_timestamp(now_ms)
                port.save_asap()  # history_last_timestamp must be persisted
        except asyncio.CancelledError:
            logger.debug("sampling task cancelled")
            break
        except Exception as e:
            logger.error("sampling task error: %s", e, exc_info=True)


async def janitor_task() -> None:
    global _pending_remove_samples

    while True:
        try:
            await asyncio.sleep(settings.core.history_janitor_interval)

            if not system.date.has_real_date_time():
                continue

            now = int(time.time())
            for port in core_ports.get_all():
                history_retention = await port.get_history_retention()
                if history_retention <= 0:
                    continue

                to_timestamp = (now - history_retention) * 1000
                logger.debug("removing old samples of %s from history", port)
                await remove_samples([port], from_timestamp=0, to_timestamp=to_timestamp)

            # Remove samples that were background-scheduled for removal
            # Group together requests w/o timestamp bounds

            ports = []
            rem_pending_remove_samples = []
            for port, from_timestamp, to_timestamp in _pending_remove_samples:
                if from_timestamp is to_timestamp is None:
                    ports.append(port)
                else:
                    rem_pending_remove_samples.append((port, from_timestamp, to_timestamp))

            if ports:
                port_ids = [p.get_id() for p in ports]
                logger.debug("removing samples of %s from history (background)", ", ".join(port_ids))
                await remove_samples(ports)

            _pending_remove_samples = rem_pending_remove_samples
            while _pending_remove_samples:
                port, from_timestamp, to_timestamp = _pending_remove_samples.pop(0)
                logger.debug("removing samples of %s from history (background)", port)
                await remove_samples([port], from_timestamp=from_timestamp, to_timestamp=to_timestamp)
        except asyncio.CancelledError:
            logger.debug("janitor task cancelled")
            break
        except Exception as e:
            logger.error("janitor task error: %s", e, exc_info=True)


def is_enabled() -> bool:
    return persist.is_samples_supported() and settings.core.history_support


async def get_samples_slice(
    port: core_ports.BasePort,
    from_timestamp: int | None = None,
    to_timestamp: int | None = None,
    limit: int | None = None,
    sort_desc: bool = False,
) -> Iterable[tuple[int, PortValue]]:
    samples = await persist.get_samples_slice(
        _PERSIST_COLLECTION, port.get_id(), from_timestamp, to_timestamp, limit, sort_desc
    )

    # Transform samples according to port type
    type_ = await port.get_type()
    integer = await port.get_attr("integer")
    samples = ((s[0], port.adapt_value_type_sync(type_, integer, s[1])) for s in samples)

    return samples


async def get_samples_by_timestamp(port: core_ports.BasePort, timestamps: list[int]) -> Iterable[GenericJSONDict]:
    now_ms = int(time.time() * 1000)
    samples_cache = _samples_cache.setdefault(port.get_id(), {})
    MISSED = {}

    results = {}
    missed_timestamps = []
    for timestamp in timestamps:
        # Look it up in cache
        sample = samples_cache.get(timestamp, MISSED)
        if sample is MISSED:
            missed_timestamps.append(timestamp)
        else:
            results[timestamp] = sample

    if missed_timestamps:
        samples = await persist.get_samples_by_timestamp(_PERSIST_COLLECTION, port.get_id(), missed_timestamps)
        samples = list(samples)

        # Transform samples according to port type
        type_ = await port.get_type()
        integer = await port.get_attr("integer")
        samples = [port.adapt_value_type_sync(type_, integer, s) for s in samples]

        for i, timestamp in enumerate(missed_timestamps):
            results[timestamp] = samples[i]

            # Add sample to cache if it's old enough
            if now_ms - timestamp > _CACHE_TIMESTAMP_MIN_AGE:
                samples_cache[timestamp] = samples[i]

    return ({"value": v, "timestamp": t} if v is not None else None for t, v in results.items())


async def save_sample(port: core_ports.BasePort, timestamp: int) -> None:
    value = port.get_last_read_value()
    if value is None:
        logger.debug("skipping null sample of %s (timestamp = %s)", port, timestamp)
        return

    logger.debug("saving sample of %s (value = %s, timestamp = %s)", port, json_utils.dumps(value), timestamp)

    await persist.save_sample(_PERSIST_COLLECTION, port.get_id(), timestamp, float(value))


async def remove_samples(
    ports: list[core_ports.BasePort],
    from_timestamp: int | None = None,
    to_timestamp: int | None = None,
    background: bool = False,
) -> int | None:
    # Invalidate samples cache for the ports
    for port in ports:
        _samples_cache.pop(port.get_id(), None)

    if background:
        for port in ports:
            _pending_remove_samples.append((port, from_timestamp, to_timestamp))
    else:
        port_ids = [p.get_id() for p in ports]
        return await persist.remove_samples(_PERSIST_COLLECTION, port_ids, from_timestamp, to_timestamp)


async def reset() -> None:
    logger.debug("clearing persisted data")
    await persist.remove_samples(_PERSIST_COLLECTION)


async def init() -> None:
    global _history_event_handler
    global _sampling_task
    global _janitor_task

    _history_event_handler = HistoryEventHandler()
    core_events.register_handler(_history_event_handler)

    _sampling_task = asyncio.create_task(sampling_task())
    _janitor_task = asyncio.create_task(janitor_task())

    await persist.ensure_index(_PERSIST_COLLECTION)


async def cleanup() -> None:
    if _sampling_task:
        _sampling_task.cancel()
        await _sampling_task

    if _janitor_task:
        _janitor_task.cancel()
        await _janitor_task



import hashlib
import logging

from datetime import datetime
from typing import Any

from tornado import httpclient

from qtoggleserver.conf import settings
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils import logging as logging_utils


logger = logging.getLogger(__name__)


class APClient(logging_utils.LoggableMixin):
    def __init__(self, mac_address: str, ip_address: str, hostname: str, moment: datetime) -> None:
        self.mac_address: str = mac_address
        self.ip_address: str = ip_address
        self.hostname: str = hostname
        self.moment: datetime = moment

        super().__init__(self.mac_address, logger)

    def __str__(self) -> str:
        return f"APClient {self.mac_address} at {self.ip_address}"

    async def request(
        self, method: str, path: str, body: Any = None, admin_password: str | None = None, no_log: bool = False
    ) -> Any:
        http_client = httpclient.AsyncHTTPClient()
        if admin_password:
            password_hash = hashlib.sha256(admin_password.encode()).hexdigest()
        else:
            password_hash = core_api_auth.EMPTY_PASSWORD_HASH

        headers = {
            "Content-Type": json_utils.JSON_CONTENT_TYPE,
            "Authorization": core_api_auth.make_auth_header(
                core_api_auth.ORIGIN_CONSUMER, username="admin", password_hash=password_hash
            ),
        }

        # TODO: this only tries standard port 80; ideally it should try connecting on a list of known ports

        timeout = settings.slaves.discover.request_timeout
        url = f"http://{self.ip_address}:80{path}"

        body_str = None
        if body is not None:
            body_str = json_utils.dumps(body)

        request = httpclient.HTTPRequest(
            url=url, method=method, headers=headers, body=body_str, connect_timeout=timeout, request_timeout=timeout
        )

        if not no_log:
            self.debug("requesting %s %s", method, path)

        try:
            response = await http_client.fetch(request, raise_error=True)
        except Exception as e:
            if not no_log:
                self.error("request %s %s failed: %s", method, path, e, exc_info=True)
            raise

        if response.body:
            return json_utils.loads(response.body)
        else:
            return None



import logging

from datetime import datetime, timezone

from .client import APClient
from .dnsmasq import DNSMasq
from .exceptions import APException
from .hostapd import HostAPD


logger = logging.getLogger(__name__)

_hostapd: HostAPD | None = None
_dnsmasq: DNSMasq | None = None


async def update() -> None:
    global _hostapd
    global _dnsmasq

    if _hostapd and not _hostapd.is_alive():
        logger.warning("hostapd died unexpectedly")
        await _hostapd.stop()
        _hostapd = None

    if _dnsmasq and not _dnsmasq.is_alive():
        logger.warning("dnsmasq died unexpectedly")
        await _dnsmasq.stop()
        _dnsmasq = None

    if None in [_hostapd, _dnsmasq]:
        # Ensure both daemons are stopped
        if _hostapd:
            logger.warning("dnsmasq is stopped but hostapd is running")
            await _hostapd.stop()
            _hostapd = None

        if _dnsmasq:
            logger.warning("hostapd is stopped but dnsmasq is running")
            await _dnsmasq.stop()
            _dnsmasq = None


def is_running() -> bool:
    return (_hostapd is not None) or (_dnsmasq is not None)


def start(
    interface: str,
    ssid: str,
    psk: str | None,
    own_ip: str,
    mask_len: int,
    start_ip: str,
    stop_ip: str,
    hostapd_binary: str | None = None,
    hostapd_cli_binary: str | None = None,
    dnsmasq_binary: str | None = None,
    hostapd_log: str | None = None,
    dnsmasq_log: str | None = None,
) -> None:
    global _hostapd
    global _dnsmasq

    logger.debug("starting")

    if is_running():
        raise APException("AP already started")

    if HostAPD.is_already_running():
        raise APException("The hostapd daemon is already running")
    if DNSMasq.is_already_running():
        raise APException("The dnsmasq daemon is already running")

    _hostapd = HostAPD(
        interface=interface,
        ssid=ssid,
        psk=psk,
        hostapd_binary=hostapd_binary,
        hostapd_cli_binary=hostapd_cli_binary,
        hostapd_log=hostapd_log,
    )

    _dnsmasq = DNSMasq(
        interface=interface,
        own_ip=own_ip,
        mask_len=mask_len,
        start_ip=start_ip,
        stop_ip=stop_ip,
        dnsmasq_binary=dnsmasq_binary,
        dnsmasq_log=dnsmasq_log,
    )

    _dnsmasq.start()
    _hostapd.start()

    logger.debug("started")


async def stop() -> None:
    global _hostapd
    global _dnsmasq

    logger.debug("stopping")

    if not is_running():
        raise APException("AP not started")

    await _hostapd.stop()
    await _dnsmasq.stop()

    _hostapd = None
    _dnsmasq = None

    logger.debug("stopped")


def get_clients() -> list[APClient]:
    if not is_running():
        raise APException("AP not started")

    leases = _dnsmasq.get_leases()

    return [
        APClient(
            mac_address=lease["mac_address"].upper(),
            ip_address=lease["ip_address"],
            hostname=lease["hostname"],
            moment=datetime.fromtimestamp(lease["timestamp"], tz=timezone.utc),
        )
        for lease in leases
    ]



from .json import JSONDriver


__all__ = ["JSONDriver"]


try:
    from .mongo import MongoDriver  # noqa: F401

    __all__.append("MongoDriver")
except ImportError:
    pass

try:
    from .postgres import PostgresDriver  # noqa: F401

    __all__.append("PostgresDriver")
except ImportError:
    pass

try:
    from .redis import RedisDriver  # noqa: F401

    __all__.append("RedisDriver")
except ImportError:
    pass



import logging
import os
import subprocess

import psutil

from qtoggleserver.conf import settings

from . import battery, conf, date, dns, fwupdate, net, storage, temperature


__all__ = [
    "battery",
    "conf",
    "date",
    "dns",
    "fwupdate",
    "net",
    "storage",
    "temperature",
]


logger = logging.getLogger(__name__)


def is_setup_mode() -> bool:
    if not settings.system.setup_mode_cmd:
        return False

    try:
        subprocess.check_call(settings.system.setup_mode_cmd, stderr=subprocess.STDOUT, shell=True)
        return True
    except subprocess.CalledProcessError:
        return False


def reboot() -> None:
    logger.debug("rebooting")
    if not settings.debug:
        os.system("reboot")


def uptime() -> int:
    with open("/proc/uptime") as f:
        line = f.readlines()[0]
        return int(float(line.split()[0]))


def get_cpu_usage() -> int:
    return int(psutil.cpu_percent())


def get_mem_usage() -> int:
    vm = psutil.virtual_memory()
    return int(100 * (1 - vm.available / vm.total))


async def init() -> None:
    logger.debug("initializing DNS")
    await dns.init()


async def cleanup() -> None:
    logger.debug("cleaning up DNS")
    await dns.cleanup()



import logging

from qtoggleserver import persist
from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList

from ..events import DashboardUpdateEvent


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_panels(request: core_api.APIRequest) -> GenericJSONList:
    return await persist.get_value("dashboard_panels", default=[])


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_panels(request: core_api.APIRequest, params: GenericJSONList) -> None:
    # core_api.validate(panels, PANELS_SCHEMA)  TODO: validate against schema

    await persist.set_value("dashboard_panels", params)

    await core_events.trigger(DashboardUpdateEvent(request=request, panels=params))


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_prefs(request: core_api.APIRequest) -> GenericJSONDict:
    prefs = await persist.get("frontend_prefs", id_=request.username) or {}
    prefs.pop("id", None)

    return prefs


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def put_prefs(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    await persist.replace("frontend_prefs", id_=request.username, record=params)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_frontend(request: core_api.APIRequest) -> GenericJSONDict:
    return {
        "prefs": sorted(await persist.query("frontend_prefs"), key=lambda p: p["id"]),
        "dashboard_panels": await persist.get_value("dashboard_panels"),
    }


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_frontend(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    # core_api.validate(panels, FRONTEND_SCHEMA)  TODO: validate against schema

    logger.debug("restoring frontend configuration")

    await persist.remove("frontend_prefs")

    prefs = params.get("prefs")
    if prefs:
        for p in prefs:
            logger.debug('restoring frontend prefs for username "%s"', p.get("id"))
            await persist.insert("frontend_prefs", p)

    dashboard_panels = params.get("dashboard_panels", [])
    logger.debug("restoring dashboard panels")
    await persist.set_value("dashboard_panels", dashboard_panels)

    # Don't supply a request to the DashboardUpdateEvent because the event isn't directly triggered by the current
    # session, but rather an indirect consequence of restoring the frontend settings. This ensures that the current user
    # will also benefit from an updated dashboard when he restores frontend from a backup.
    await core_events.trigger(DashboardUpdateEvent(panels=dashboard_panels))



import re

from collections.abc import Callable
from typing import Any

import jsonschema

from qtoggleserver.core.typing import GenericJSONDict

from . import APIError


POST_RESET = {
    "type": "object",
    "properties": {"factory": {"type": "boolean"}},
    "additionalProperties": False,
}

PATCH_FIRMWARE = {
    "type": "object",
    "properties": {"version": {"type": "string", "maxLength": 256}, "url": {"type": "string", "maxLength": 256}},
    "additionalProperties": False,
    "anyOf": [{"required": ["version"]}, {"required": ["url"]}],
}

PUT_PORTS = {"type": "array", "items": {"type": "object"}, "additionalProperties": True}

POST_PORTS = {
    "type": "object",
    "properties": {
        "id": {"type": "string", "pattern": "^[a-zA-Z_][a-zA-Z0-9_.-]{0,63}$"},
        "type": {"enum": ["boolean", "number"]},
        "min": {"type": "number"},
        "max": {"type": "number"},
        "integer": {"type": "boolean"},
        "step": {"type": "number"},
        "choices": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "value": {"oneOf": [{"type": "boolean"}, {"type": "number"}]},
                    "display_name": {"type": "string", "maxLength": 64},
                },
                "required": ["value"],
                "additionalProperties": False,
            },
            "minItems": 2,
            "maxItems": 256,
        },
    },
    "additionalProperties": False,
    "required": ["id", "type"],
}

PATCH_PORT_SEQUENCE = {
    "type": "object",
    "properties": {
        "values": {
            "type": "array",
            "items": {"oneOf": [{"type": "boolean"}, {"type": "number"}]},
            "minItems": 0,
            "maxItems": 256,
        },
        "delays": {
            "type": "array",
            "items": {"type": "integer", "min": 0, "max": 60000},
            "minItems": 0,
            "maxItems": 256,
        },
        "repeat": {"type": "integer", "min": 0, "max": 65535},
    },
    "additionalProperties": False,
    "required": ["values", "delays", "repeat"],
}

PATCH_WEBHOOKS = {
    "type": "object",
    "properties": {
        "enabled": {"type": "boolean"},
        "scheme": {"enum": ["http", "https"]},
        "host": {"type": "string", "maxLength": 256},
        "port": {"type": "integer", "min": 1, "max": 65535},
        "path": {"type": "string", "maxLength": 256},
        "password": {"type": "string", "maxLength": 64},
        "password_hash": {"type": "string", "pattern": "^[a-f0-9]{64}$"},
        "timeout": {"type": "integer", "minimum": 1, "maximum": 3600},
        "retries": {"type": "integer", "minimum": 0, "maximum": 10},
    },
    "additionalProperties": False,
    "required": ["enabled", "scheme", "host", "port", "path", "timeout", "retries"],
}

PATCH_REVERSE = {
    "type": "object",
    "properties": {
        "enabled": {"type": "boolean"},
        "scheme": {"enum": ["http", "https"]},
        "host": {"type": "string", "maxLength": 256},
        "port": {"type": "integer", "min": 1, "max": 65535},
        "path": {"type": "string", "maxLength": 256},
        "password": {"type": "string", "maxLength": 64},
        "password_hash": {"type": "string", "pattern": "^[a-f0-9]{64}$"},
        "device_id": {"type": "string", "pattern": "^[a-zA-Z0-9]{1,64}"},
        "timeout": {"type": "integer", "minimum": 1, "maximum": 3600},
    },
    "additionalProperties": False,
    "required": ["enabled", "scheme", "host", "port", "path", "device_id", "timeout"],
}

POST_INTROSPECT = {
    "type": "object",
    "properties": {"imports": {"type": "array", "items": {"type": "string"}}, "code": {"type": "string"}},
    "additionalProperties": False,
    "required": ["code"],
}


def _validate_schema(json: Any, schema: GenericJSONDict) -> tuple[str, str | None] | None:
    try:
        jsonschema.Draft4Validator(schema=schema).validate(json)
        return None
    except jsonschema.ValidationError as e:
        try:
            field = e.path.pop()
            return "invalid", field
        except Exception:
            try:
                field = re.match(r"\'(\w+)\' is a required property", e.message).group(1)
                return "missing", field
            except Exception:
                try:
                    field = re.match(r"Additional properties are not allowed \(u?\'(\w+)\'.*", e.message).group(1)
                    return "unexpected", field
                except Exception:
                    return "invalid", None


def validate(
    params: Any,
    schema: GenericJSONDict,
    invalid_field_code: str | Callable = "invalid-field",
    unexpected_field_code: str | Callable = "invalid-request",
    missing_field_code: str | Callable = "missing-field",
    invalid_request_code: str = "invalid-request",
    invalid_field_name: str = "field",
    unexpected_field_name: str = "field",
    missing_field_name: str = "field",
) -> None:
    validation_error = _validate_schema(params, schema)
    if validation_error:
        error, field = validation_error
        if error == "invalid":
            if field:
                if callable(invalid_field_code):
                    invalid_field_code = invalid_field_code(field)

                raise APIError(400, invalid_field_code, **{invalid_field_name: field})
            else:
                raise APIError(400, invalid_request_code)
        elif error == "missing":
            if callable(missing_field_code):
                missing_field_code = missing_field_code(field)

            raise APIError(400, missing_field_code, **{missing_field_name: field})
        elif error == "unexpected":
            if callable(unexpected_field_code):
                unexpected_field_code = unexpected_field_code(field)

            raise APIError(400, unexpected_field_code, **{unexpected_field_name: field})
        else:
            raise APIError(400, "invalid-request")



import logging

from qtoggleserver.core import api as core_api
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.system import fwupdate


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_firmware(request: core_api.APIRequest) -> GenericJSONDict:
    current_version = await fwupdate.get_current_version()
    status = await fwupdate.get_status()

    if status in (fwupdate.STATUS_IDLE, fwupdate.STATUS_ERROR):
        try:
            latest_version, latest_date, latest_url = await fwupdate.get_latest()

            return {
                "version": current_version,
                "latest_version": latest_version,
                "latest_date": latest_date,
                "latest_url": latest_url,
                "status": status,
            }
        except Exception as e:
            logger.error("get latest firmware failed: %s", e, exc_info=True)

            return {"version": current_version, "status": status}
    else:
        return {"version": current_version, "status": status}


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_firmware(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, core_api_schema.PATCH_FIRMWARE)

    status = await fwupdate.get_status()
    if status not in (fwupdate.STATUS_IDLE, fwupdate.STATUS_ERROR):
        raise core_api.APIError(503, "busy")

    if params.get("url"):
        await fwupdate.update_to_url(params["url"])
    else:  # assuming params['version']
        await fwupdate.update_to_version(params["version"])



import logging

from qtoggleserver import peripherals
from qtoggleserver.core import api as core_api
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList
from qtoggleserver.peripherals.api import schema as peripherals_api_schema


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_peripherals(request: core_api.APIRequest) -> GenericJSONList:
    return [p.to_json() for p in peripherals.get_all()]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_peripherals(request: core_api.APIRequest, params: GenericJSONDict) -> GenericJSONDict:
    core_api_schema.validate(params, peripherals_api_schema.POST_PERIPHERALS)

    name = params.get("name")
    if name and peripherals.get(name):
        raise core_api.APIError(400, "duplicate-peripheral")

    try:
        peripheral = await peripherals.add(params)
    except peripherals.NoSuchDriver:
        raise core_api.APIError(404, "no-such-driver")
    except peripherals.DuplicatePeripheral:
        raise core_api.APIError(400, "duplicate-peripheral")
    except TypeError as e:
        if "arguments" in str(e):
            raise core_api.APIError(400, "invalid-field", field="params")
        else:
            raise core_api.APIError(400, "invalid-request", details=str(e))
    except Exception as e:
        raise core_api.APIError(400, "invalid-request", details=str(e))

    try:
        await peripheral.init_ports()
    except Exception as e:
        await peripherals.remove(peripheral.get_id())
        raise core_api.APIError(400, "invalid-request", details=str(e))

    return peripheral.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_peripheral(request: core_api.APIRequest, peripheral_id: str) -> None:
    p = peripherals.get(peripheral_id)
    if not p:
        raise core_api.APIError(404, "no-such-peripheral")
    if p.is_static():
        raise core_api.APIError(400, "peripheral-not-removable")

    await p.cleanup_ports(persisted_data=True)
    await peripherals.remove(peripheral_id, persisted_data=True)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_peripherals(request: core_api.APIRequest, params: GenericJSONList) -> None:
    core_api_schema.validate(params, peripherals_api_schema.PUT_PERIPHERALS)

    logger.debug("restoring peripherals")

    for p in peripherals.get_all():
        if p.is_static():
            continue
        await p.cleanup_ports(persisted_data=True)
        await peripherals.remove(p.get_id(), persisted_data=True)

    peripheral_list = []
    for par in params:
        if par.pop("static", None):
            continue
        p = await peripherals.add(par)
        peripheral_list.append(p)

    for p in peripheral_list:
        await p.init_ports()



import logging

import psutil

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd


_has_sensors_battery = None

logger = logging.getLogger(__name__)


class BatteryError(Exception):
    pass


def has_battery_support() -> bool:
    global _has_sensors_battery

    if _has_sensors_battery is None:
        try:
            _ = psutil.sensors_battery().percent
            _has_sensors_battery = True
            logger.debug("battery sensor support detected")
        except AttributeError:
            _has_sensors_battery = False

    return _has_sensors_battery or bool(settings.system.battery.get_cmd)


def get_battery_level() -> int | None:
    if settings.system.battery.get_cmd:
        return int(
            run_get_cmd(
                settings.system.battery.get_cmd, cmd_name="battery", exc_class=BatteryError, required_fields=["level"]
            )["level"]
        )

    if _has_sensors_battery:
        return int(psutil.sensors_battery().percent)
    else:
        return None



import logging

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONList
from qtoggleserver.system import conf as system_conf


logger = logging.getLogger(__name__)


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_backup_endpoints(request: core_api.APIRequest) -> GenericJSONList:
    endpoints = [{"path": "/peripherals", "display_name": "Peripherals", "restore_method": "PUT", "order": 15}]

    if system_conf.can_write_conf_file():
        endpoints.append(
            {"path": "/system", "display_name": "System Configuration", "restore_method": "PUT", "order": 5}
        )

    if settings.frontend.enabled:
        endpoints.append(
            {"path": "/frontend", "display_name": "App Configuration", "restore_method": "PUT", "order": 45}
        )

    return endpoints



import subprocess

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd


WIFI_RSSI_EXCELLENT = -50
WIFI_RSSI_GOOD = -60
WIFI_RSSI_FAIR = -70


class NetError(Exception):
    pass


def has_ip_support() -> bool:
    return bool(settings.system.net.ip.get_cmd and settings.system.net.ip.set_cmd)


def get_ip_config() -> dict[str, str]:
    assert settings.system.net.ip.get_cmd

    return run_get_cmd(
        settings.system.net.ip.get_cmd,
        cmd_name="IP config",
        exc_class=NetError,
        required_fields=[
            "address",
            "netmask",
            "gateway",
            "dns",
            "address_current",
            "netmask_current",
            "gateway_current",
            "dns_current",
        ],
    )


def set_ip_config(address: str, netmask: str, gateway: str, dns: str) -> None:
    assert settings.system.net.ip.set_cmd

    run_set_cmd(
        settings.system.net.ip.set_cmd,
        cmd_name="IP config",
        exc_class=NetError,
        address=address,
        netmask=netmask,
        gateway=gateway,
        dns=dns,
    )


def reset_ip_config() -> None:
    set_ip_config(address="", netmask="", gateway="", dns="")


def has_wifi_support() -> bool:
    return bool(settings.system.net.wifi.get_cmd and settings.system.net.wifi.set_cmd)


def get_wifi_config() -> dict[str, str]:
    assert settings.system.net.wifi.get_cmd

    result = run_get_cmd(
        settings.system.net.wifi.get_cmd,
        cmd_name="Wi-Fi config",
        log_values=False,
        exc_class=NetError,
        required_fields=["ssid", "psk", "bssid", "bssid_current", "rssi_current"],
    )

    if result["rssi_current"]:
        rssi = int(result["rssi_current"])
        if rssi >= WIFI_RSSI_EXCELLENT:
            strength = 3
        elif rssi >= WIFI_RSSI_GOOD:
            strength = 2
        elif rssi >= WIFI_RSSI_FAIR:
            strength = 1
        else:
            strength = 0

        result["signal_strength_current"] = str(strength)
    else:
        result["signal_strength_current"] = "0"

    return result


def set_wifi_config(ssid: str, psk: str, bssid: str) -> None:
    assert settings.system.net.wifi.set_cmd

    run_set_cmd(
        settings.system.net.wifi.set_cmd,
        cmd_name="Wi-Fi config",
        exc_class=NetError,
        log_values=False,
        ssid=ssid,
        psk=psk,
        bssid=bssid,
    )


def reset_wifi_config() -> None:
    set_wifi_config(ssid="", psk="", bssid="")


def get_default_interface() -> str | None:
    try:
        output = subprocess.check_output(["ip", "route"]).decode().strip()
    except Exception as e:
        raise NetError("Could not determine default route") from e

    lines = output.split("\n")
    if not lines:
        raise NetError("Could not determine default route")

    line = next((line for line in lines if line.count("default via")), None)
    if not line:
        return None

    return line.split()[4]  # the 5th part is the interface



from __future__ import annotations

import logging
import queue

from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPResponse

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)

_webhooks: Webhooks | None = None


# TODO: add password support


class WebhooksError(Exception):
    pass


class InvalidParamError(WebhooksError):
    def __init__(self, param: str) -> None:
        self.param: str = param

        super().__init__(f"Invalid field: {param}")


class WebhooksRequest:
    def __init__(self, body: GenericJSONDict) -> None:
        self.body: GenericJSONDict = body
        self.retries: int = 0


class Webhooks:
    def __init__(
        self,
        scheme: str | None = None,
        host: str | None = None,
        port: int | None = None,
        path: str | None = None,
        timeout: int | None = None,
        retries: int | None = None,
        **kwargs,
    ) -> None:
        # The enabled value comes with kwargs but is ignored; webhooks will be explicitly enabled afterwards

        self._enabled: bool = False
        self._scheme: str | None = scheme
        self._host: str | None = host
        self._port: int | None = port
        self._path: str | None = path
        self._timeout: int | None = timeout
        self._retries: int | None = retries
        self._url: str | None = None

        self._queue: queue.Queue = queue.Queue(settings.core.event_queue_size)

    def __str__(self) -> str:
        s = "webhooks"

        if not self._enabled:
            s += " [disabled]"

        if self._scheme:
            s += " " + self.get_url()

        return s

    def is_enabled(self) -> bool:
        return self._enabled

    def enable(self) -> None:
        if self._enabled:
            return

        logger.debug("enabling %s", self)

        self._enabled = True

    def disable(self) -> None:
        if not self._enabled:
            return

        logger.debug("disabling %s", self)

        self._enabled = False

        # Drop all queued requests
        while not self._queue.empty():
            self._queue.get()

    def get_url(self) -> str:
        if not self._url:
            if self._scheme == "http" and self._port == 80 or self._scheme == "https" and self._port == 443:
                self._url = f"{self._scheme}://{self._host}{self._path}"
            else:
                self._url = f"{self._scheme}://{self._host}:{self._port}{self._path}"

        return self._url

    def call(self, body: GenericJSONDict) -> None:
        if not self._enabled:
            return None

        request = WebhooksRequest(body)

        if self._queue.empty():
            return self._request(request)

        try:
            self._queue.put(request)
        except queue.Full:
            logger.error("%s: queue is full", self)

        return None

    def _request(self, request: WebhooksRequest) -> None:
        if not self._enabled:
            return

        url = self.get_url()
        body = json_utils.dumps(request.body)

        def on_response(response: HTTPResponse) -> None:
            try:
                core_responses.parse(response)
            except core_responses.Error as e:
                logger.error("%s: call failed: %s", self, e)

                # Retry mechanism
                if not self._retries:
                    return self._check_pending()

                if request.retries <= self._retries:
                    request.retries += 1
                    logger.debug("%s: resending request (retry %s/%s)", self, request.retries, self._retries)

                    self._request(request)
                else:
                    self._check_pending()
            else:
                logger.debug("%s: call succeeded", self)
                self._check_pending()

            return None

        http_client = AsyncHTTPClient()
        # TODO use webhooks password
        headers = {
            "Content-Type": json_utils.JSON_CONTENT_TYPE,
            "Authorization": core_api_auth.make_auth_header(
                core_api_auth.ORIGIN_DEVICE, username=None, password_hash=core_device_attrs.normal_password_hash
            ),
        }
        request = HTTPRequest(
            url,
            "POST",
            headers=headers,
            body=body,
            connect_timeout=self._timeout,
            request_timeout=self._timeout,
            follow_redirects=True,
        )

        logger.debug("%s: calling", self)

        http_client.fetch(request, callback=on_response)  # TODO implement me using await

    def _check_pending(self) -> None:
        try:
            request = self._queue.get(block=False)
        except queue.Empty:
            return

        self._request(request)

    def to_json(self) -> GenericJSONDict:
        d = {
            "enabled": self._enabled,
            "scheme": self._scheme,
            "host": self._host,
            "port": self._port,
            "path": self._path,
            "timeout": self._timeout,
            "retries": self._retries,
        }

        return d


def get() -> Webhooks | None:
    return _webhooks


def setup(
    enabled: bool,
    scheme: str | None = None,
    host: str | None = None,
    port: int | None = None,
    path: str | None = None,
    timeout: int | None = None,
    retries: int | None = None,
    **kwargs,
) -> None:
    global _webhooks

    if _webhooks and _webhooks.is_enabled():
        _webhooks.disable()

    _webhooks = Webhooks(scheme, host, port, path, timeout, retries)
    if enabled:
        _webhooks.enable()


async def load() -> None:
    data = await persist.get_value("webhooks")
    if data is None:
        setup(enabled=False)
        logger.debug("loaded %s", _webhooks)

        return

    setup(**data)
    logger.debug("loaded %s", _webhooks)


async def save() -> None:
    if _webhooks is None:
        return

    logger.debug("saving persisted data")
    await persist.set_value("webhooks", _webhooks.to_json())


async def reset() -> None:
    logger.debug("clearing persisted data")
    await persist.remove("webhooks")


async def init() -> None:
    logger.debug("loading persisted data")
    await load()


async def cleanup() -> None:
    pass



import psutil

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd


class TemperatureError(Exception):
    pass


def has_temperature_support() -> bool:
    return bool(settings.system.temperature.sensor_name) or bool(settings.system.temperature.get_cmd)


def get_temperature() -> int | None:
    if settings.system.temperature.get_cmd:
        return int(
            run_get_cmd(
                settings.system.temperature.get_cmd,
                cmd_name="temperature",
                exc_class=TemperatureError,
                required_fields=["value"],
            )["value"]
        )

    if settings.system.temperature.sensor_name:
        temperatures = psutil.sensors_temperatures()
        temp_info_list = temperatures[settings.system.temperature.sensor_name]
        temp_info = temp_info_list[settings.system.temperature.sensor_index]

        return int(temp_info.current)
    else:
        return None
